import unittest
import sys
sys.path.append('../../')
import numpy as np
import time

from verstack import Multicore

iterable = np.random.randint(0, 100, 10000)

# test overall Multicore not broken

def func_to_parallel(x):
    print('\nsleeping 3 sec\n')
    time.sleep(3)
    return(x**2)

class TestMulticore(unittest.TestCase):
    
    
    def test_Multicore(self):
        worker = Multicore(workers = 2)
        multicore_result = worker.execute(func_to_parallel, iterable)
        manual_result = iterable**2
        result = np.all(multicore_result==manual_result)
        self.assertTrue(result)
        
if __name__ == '__main__':
    unittest.main()


import unittest
import sys
sys.path.append('../../')
from common import generate_data

from verstack import NaNImputer

# test overall DateParser not being broken

class TestNaNImputer(unittest.TestCase):
    
    def test_NaNImputer(self):
        df = generate_data()
        module = NaNImputer()
        transformed = module.impute(df)
        result = sum(transformed.isnull().sum()) == 0
        self.assertTrue(result)
        
if __name__ == '__main__':
    unittest.main()


import pandas as pd
import numpy as np
from verstack.categoric_encoders.args_validators import is_bool_na_sentinel, assert_fit_transform_args, assert_transform_args, assert_binary_target
from verstack.tools import Printer

class WeightOfEvidenceEncoder():
    '''
    Encoder to represent categoric variables by Weight of evidence in regards to the binary target variable.
    If encoded value is negative - it represents a category that is more heavily enclided to the negative target class (0).
    
    When fit_transform() is used on a train set, variable is encoded with adding minor noize to reduce the risk of overfitting.
    When transform() unseen categories are encoded by 0 WoE.
    '''

    __version__ = '0.1.1'    
    
    def __init__(self, na_sentinel = True, **kwargs):
        '''
        Initialize WeightOfEvidence instance.
        
        Parameters
        ----------
        na_sentinel : bool, optional
            If True - encode NaN by 0 WoE, else keep NaN untransformed.
            The default is True.
        kwargs : category_encoders.woe.WOEEncoder parameters can be found at:
            https://contrib.scikit-learn.org/category_encoders/woe.html

        Returns
        -------
        None.
        '''
        self.printer = Printer(verbose=True)
        self._na_sentinel = is_bool_na_sentinel(na_sentinel)
        self._colname = None
        self._params = {'randomized':True, 'random_state':42, 'handle_missing':'return_nan' if not na_sentinel else 'value'} if not kwargs else kwargs
        if 'random_state' not in self._params.keys():
            self._params['random_state'] = 42
        if 'randomized' not in self._params.keys():
            self.randomized = True
        self.__generic_encoder = None # category_encoders.woe.WOEEncoder placeholder
        self.__pattern = None # for inverse_transform without noize

    def __repr__(self): # what an object will say when you call/print it
        return 'verstack.categoric_encoders.WeightOfEvidenceEncoder'
        
    @property 
    def colname(self): 
        return self._colname

    @property 
    def na_sentinel(self): 
        return self._na_sentinel

    @property 
    def params(self): 
        return self._params

    def fit_transform(self, df, colname, targetname):
        '''
        Fit encoder, transform column in df, save attributes for transform(/inverse_transform().
                                                                           
        Variable is encoded with adding minor noize to reduce the risk of overfitting.

        Parameters
        ----------
        df : pd.DataFrame
            Data containing the colname to transform.
        colname : str
            Column name in df to be transformed.
        targetname : str
            column name for extracting the mean values for each colname category.

        Returns
        -------
        transformed_df : pd.DataFrame
            Data with the column transformed.
        '''
        
        assert_fit_transform_args(df, colname)
        assert_binary_target(df, targetname)
        encoded_df = df.copy()
        self._colname = colname
        from category_encoders import WOEEncoder
        generic_encoder = WOEEncoder(**self._params)
        encoded_column = generic_encoder.fit_transform(df[colname], df[targetname])
        self.__generic_encoder = generic_encoder
        encoded_df[self._colname] = encoded_column
        
        # save inverse_transform pattern for test set (without noize)
        woe_vals_no_noize = self.transform(df)[self._colname].unique()
        original_vals = df[self._colname].unique()
        self.__pattern = dict(zip(woe_vals_no_noize, original_vals))

        return encoded_df        
        
    def transform(self, df):
        '''
        WOE Encode data column saved in self._colname using saved patterns.
        Unseen categories are encoded by 0 WoE.
        
        Parameters
        ----------
        df : pd.DataFrame
            Data containing colname that needs encoding by initial class instance.

        Returns
        -------
        df : pd.DataFrame
            Data containing the transformed column.
        '''
        
        assert_transform_args(df)
        encoded_df = df.copy()
        encoded_column = self.__generic_encoder.transform(df[self._colname])
        encoded_df[self._colname] = encoded_column
        return encoded_df

    def _check_random_state(self, seed):
        """Turn seed into a np.random.RandomState instance (used for inverse_transform()).
    
        Parameters
        ----------
        seed : None, int or instance of RandomState
            If seed is None, return the RandomState singleton used by np.random.
            If seed is an int, return a new RandomState instance seeded with seed.
            If seed is already a RandomState instance, return it.
            Otherwise raise ValueError.
        """
        import numbers
        if seed is None or seed is np.random:
            return np.random.mtrand._rand
        if isinstance(seed, numbers.Integral):
            return np.random.RandomState(seed)
        if isinstance(seed, np.random.RandomState):
            return seed
        raise ValueError('%r cannot be used to seed a numpy.random.RandomState'
                         ' instance' % seed)

    def inverse_transform(self, df):       
        '''
        Return transformed column in df to original values.

        Parameters
        ----------
        df : pd.DataFrame
            Data containing the column which had been passed at fit_transform().

        Returns
        -------
        transformed_df : pd.DataFrame
            Data containing the transformed column.
        '''
        
        assert_transform_args(df)
        # first check for inverse_transform without noize
        if np.all([x in list(self.__pattern.keys()) for x in df[self._colname].unique()]):
            self.printer.print('Inverse transforming without noize', order=3)
            inverse_encoded_col = df[self._colname].map(self.__pattern)
            inverse_encoded_df = df.copy()
            inverse_encoded_df[self._colname] = inverse_encoded_col
            return inverse_encoded_df
        else:
            inverse_encoded_df = df.copy()
            random_state_generator = self._check_random_state(self.__generic_encoder.random_state)
            inverse_encoded_col = df[self._colname]  / random_state_generator.normal(1., self.__generic_encoder.sigma, df[self._colname].shape[0])
            
            ordinal_string_mapping = self.__generic_encoder.ordinal_encoder.mapping[0]['mapping']
            ordinal_string_mapping = {val: key for key, val in ordinal_string_mapping.items()}
            ordinal_woe_mapping = self.__generic_encoder.mapping[self._colname]
            final_mapping = {val_ord_str: ordinal_woe_mapping[key_ord_str] for key_ord_str, val_ord_str in ordinal_string_mapping.items()}
            final_mapping = {val: key for key, val in final_mapping.items()}
            # decrease the float depth for mapping (some division results in inverse_encoded_col might have slight differences)
            inverse_encoded_df[self._colname] = inverse_encoded_col.astype('float32').map({np.float32(key):val for key, val in final_mapping.items()})
            return inverse_encoded_df




import pandas as pd

def validate_objective(value):
    if not isinstance(value, str): 
        raise TypeError('objective must be a string')
    if value not in ['regression', 'binary', 'multiclass']:
        raise NameError('objective can take values: "regression", "binary", "multiclass"')

def validate_bool_arg(value):
    '''Validate arguments "auto", "meta_feats", "verbose"'''
    if not isinstance(value, bool):
        raise TypeError('auto must be bool')

def validate_num_auto_layers(value):
    if not isinstance(value, int): 
        raise TypeError('num_auto_layers must be an integer')
    if value not in [1,2]:
        raise ValueError('num_auto_layers can take values: 1, 2')

def validate_epochs(value):
    if not isinstance(value, int): 
        raise TypeError('epochs must be an integer')
    if value < 1:
        raise ValueError('epochs must be a positive integer')

def validate_gridsearch_iterations(value):
    if not isinstance(value, int): 
        raise TypeError('gridsearch_iterations must be an integer')
    if value < 0:
        raise ValueError('gridsearch_iterations must be >=0')

def validate_stacking_feats_depth(value):
    if not isinstance(value, int): 
        raise TypeError('stacking_feats_depth must be an integer')
    if value not in 4 in range(5):
        raise ValueError('stacking_feats_depth can take values: 1, 2, 3, 4')

def validate_models_list(value):
    '''Validate add_layer(models_list)'''

    def validate_class_instance(value):
        '''Validate model in models_list is a class instance'''
        if not hasattr(value, '__dict__'):
            print(f'{value} is not a valid model, it must be a class instance')

    if not isinstance(value, list):
        raise TypeError('models_list must be a list')
    if len(value) < 1:
        raise ValueError('models_list must include at least 1 model')
    for val in value:
        validate_class_instance(val)
        
def validate_fit_transform_args(X, y):
    if not isinstance(X, pd.DataFrame):
        raise TypeError('X must be a pandas.DataFrame')
    if not isinstance(y, pd.Series):
        raise TypeError('y must be a pandas.Series')
    if not len(X) == len(y):
        raise ValueError('X and y must same length')

def validate_transform_args(X):
    if not isinstance(X, pd.DataFrame):
        raise TypeError('X must be a pandas.DataFrame')



import unittest
import sys
sys.path.append('../../')
from common import generate_data

from verstack import Stacker

# test overall LGBMTuner not being broken

class TestDateLGBMTuner(unittest.TestCase):
    
    def test_LGBMTuner(self):
        
        df = generate_data(processed=True)
        module = Stacker('regression', auto=True, epochs=5, gridsearch_iterations=4)
        X = df.drop('y', axis = 1)
        y = df['y']
        X_train = module.fit_transform(X, y)
        result_layers_with_feats_present = len(module.stacked_features)>0
        self.assertTrue(result_layers_with_feats_present)
        
if __name__ == '__main__':
    unittest.main()


import pandas as pd
import numpy as np

def validate_features_argument(value):
    if not isinstance(value, pd.DataFrame): 
        raise TypeError('X (features) must be a pandas DataFrame')
    if len(value) == 0:
        raise Exception('X (features) must contain data')

def validate_target_argument(value):
    if not isinstance(value, pd.Series):
        raise TypeError('y (target variable) must be a pandas Series')
    if len(value) == 0:
        raise Exception('X (features) must contain data')

def validate_threshold_argument(value):
    if not isinstance(value, float):
        raise TypeError('threshold must be a float')
    if not 0 < value < 1:
        raise ValueError('threshold must can take values 0 < threshold < 1')
    
def validate_plotting_interactive_argument(value):
    if not isinstance(value, bool):
        raise TypeError('interactive argument can be True/False')

def validate_plotting_legend_argument(value):
    if not isinstance(value, bool):
        raise TypeError('legent argument can be True/False')

def validate_plot_importances_n_features_argument(value):
    if not isinstance(value, int):
        raise TypeError('n_features argument must be of type int')        
    if value <= 0:
        raise ValueError('n_features argument must be positive')
        
def validate_plot_importances_figsize_argument(value):
    if not isinstance(value, tuple):
        raise TypeError('figsize argument must be a tuple. E.g. (15,10)')
    if len(value) != 2:
        raise ValueError('figsize must contain two values. E.g. (15,10)')
    if np.any([val <= 0 for val in value]):
        raise ValueError('figsize must contain two positive values. E.g. (15,10)')

def validate_numpy_ndarray_arguments(value):
    if not isinstance(value, np.ndarray):
        raise TypeError('Arguments to fit_optimized() must be of type numpy.array')

def validate_params_argument(value):
    if not isinstance(value, dict):
        raise TypeError('params argument must be of type dict')

def validate_verbose_eval_argument(value):
    if not isinstance(value, int):
        raise TypeError('verbose_eval argument must be of type int')
    if value <= 0:
        raise ValueError('verbose_eval must be positive')



import pandas as pd
import numpy as np
from verstack.categoric_encoders.args_validators import is_bool_na_sentinel, assert_fit_transform_args, assert_transform_args
from verstack.tools import Printer

class OneHotEncoder():
    '''
    Encoder to represent categoric (multiclass) variable as a set of 
    binary variables one for each class.
    
    '''
    __version__ = '0.1.2'
    
    def __init__(self, na_sentinel = True):
        '''
        Initialize OneHotEncoder instance.
        
        Parameters
        ----------
        na_sentinel : bool, optional
            If True - create a separate column for np.nan values 
            else will not create a column for np.nan values. The default is True.
            
        Returns
        -------
        None.
        '''
        self.printer = Printer(verbose=True)
        self._categories = None # save for inverse_transform()
        self._colname = None
        self._na_sentinel = is_bool_na_sentinel(na_sentinel)
        self.__one_hot_cols_sequence = None # save for transform()
        self.__original_fit_transform_cols_sequence = None # save for inverse_transform()
        self.__original_transform_cols_sequence = None # save for inverse_transform()
        self.__prefix = None

    def __repr__(self): # what an object will say when you call/print it
        return 'verstack.categoric_encoders.OneHotEncoder'

    # ----------------------------------------------------------------------------------------
    # define getters
    @property 
    def categories (self): 
        return self._categories

    @property 
    def colname(self): 
        return self._colname

    @property 
    def na_sentinel(self): 
        return self._na_sentinel
    # ----------------------------------------------------------------------------------------


    def _create_output_df(self, df, one_hot_df):
        '''Common method for concatenating the one_hot_df to original df'''
        resulting_df = pd.concat([df, one_hot_df], axis = 1)
        # drop original encoded column
        resulting_df.drop(self._colname, axis = 1, inplace = True)
        return resulting_df
    
    def fit_transform(self, df, colname, prefix = None):
        '''
        Fit encoder, transform column in df, save attributes for transform(/inverse_transform().

        Change na_sentinel to False if df[colname].isnull().sum() == 0

        Parameters
        ----------
        df : pd.DataFrame
            Data containing the colname to transform.
        colname : str
            Column name in df to be transformed.
        prefix : str/int/float/bool/none, opnional.
            String to append DataFrame column names.
            The default is None.

        Returns
        -------
        transformed_df : pd.DataFrame
            Data with the column transformed.
        '''
        
        assert_fit_transform_args(df, colname)
        
        # change na_sentinel if col does not include NaNs
        if df[colname].isnull().sum() == 0:
            self._na_sentinel = False

        self._colname = colname
        self.__original_fit_transform_cols_sequence = df.columns.tolist()
        self.__prefix = prefix
        data = pd.Series(df[colname])
        one_hot_df = pd.get_dummies(df[colname], prefix = self.__prefix, dummy_na = self._na_sentinel)
        self._categories = df[colname].unique().tolist()
        if not self._na_sentinel:
            self._categories = [x for x in self._categories if x == x]
        self.__one_hot_cols_sequence = one_hot_df.columns.tolist()
        return self._create_output_df(df, one_hot_df)

    def _align_to_fit_transform(self, transformed):
        '''
        Correct columns order of transformed df to the columns order recorded
        during fit_transform

        Parameters
        ----------
        transformed : pd.DataFrame
            Dataframe after transform.

        Returns
        -------
        transformed : pd.DataFrame
            Dataframe after columns order correction.

        '''
        if transformed.columns.tolist() == self.__one_hot_cols_sequence:
            return transformed
        else:
            for col in self.__one_hot_cols_sequence:
                if col not in transformed:
                    self.printer.print(f'added empty one_hot column {col}', order=3)
                    transformed[col] = 0
            transformed = transformed[self.__one_hot_cols_sequence]
        return transformed

    def transform(self, df):
        '''
        One-hot-encode data column saved in self.colname using saved patterns

        Parameters
        ----------
        df : pd.DataFrame
            Data containing the column which had been passed at fit_transform().

        Returns
        -------
        transformed_df : pd.DataFrame
            Data containing a set of one-hot-encoded columns, without the original column.

        '''

        assert_transform_args(df)
        self.__original_transform_cols_sequence = df.columns.tolist()        
        one_hot_df = pd.get_dummies(df[self._colname], prefix=self.__prefix, dummy_na = self._na_sentinel)
        one_hot_df = self._align_to_fit_transform(one_hot_df)
        return self._create_output_df(df, one_hot_df)

    def inverse_transform(self, df):
        '''
        Return transformed column in df to original values.

        Parameters
        ----------
        df : pd.DataFrame
            Data containing the column which had been passed at fit_transform().

        Returns
        -------
        transformed_df : pd.DataFrame
            Data containing the transformed column.
        '''
        
        assert_transform_args(df)
        try: # check if the one-hot-encoded columns are present in df
            assert np.all([col in df for col in self.__one_hot_cols_sequence]) 
        except AssertionError:
            self.printer.print('Passed data does not contain the ohe-hot-encoded columns sequence', order='error')
            return df        
        inverse_transform_map = dict(zip(self.__one_hot_cols_sequence, self._categories))
        inverse_transformed_col = df[self.__one_hot_cols_sequence].replace(0, np.nan).idxmax(axis=1)
        inverse_transformed_col = inverse_transformed_col.map(inverse_transform_map)
        resulting_df = df.drop(self.__one_hot_cols_sequence, axis = 1)
        resulting_df[self._colname] = inverse_transformed_col
        try:
            resulting_df = resulting_df[self.__original_fit_transform_cols_sequence]
        except KeyError:
            resulting_df = resulting_df[self.__original_transform_cols_sequence]
        return resulting_df



# standalone classes
from verstack.NaNImputer import NaNImputer
from verstack.Multicore import Multicore
from verstack.ThreshTuner import ThreshTuner
from verstack.DateParser import DateParser
from verstack.FeatureSelector import FeatureSelector
# categoric_encoders imports
from verstack.categoric_encoders.Factorizer import Factorizer
from verstack.categoric_encoders.OneHotEncoder import OneHotEncoder
from verstack.categoric_encoders.MeanTargetEncoder import MeanTargetEncoder
from verstack.categoric_encoders.FrequencyEncoder import FrequencyEncoder
from verstack.categoric_encoders.WeightOfEvidenceEncoder import WeightOfEvidenceEncoder
# LGBMTuner import
from verstack.lgbm_optuna_tuning.LGBMTuner import LGBMTuner
# Stacker import
from verstack.stacking.Stacker import Stacker
#create a __version__ attribute in the verstack class
from .version import __version__



import pandas as pd
import numpy as np
from sklearn.feature_selection import RFECV
from verstack.tools import Printer
from verstack.tools import timer
# SUBSET DATA, AND MAYBE OTHERS MOVE TO nova.common.py script

class FeatureSelector:  

    __version__ = '0.0.4'

    def __init__(self, **kwargs):
                 # objective = 'regression', 
                 # auto = False,
                 # allowed_score_gap = 0.0,
                 # final_scoring_model = None,
                 # default_model_linear = False,
                 # custom_model = None, 
                 # subset_size_mb = 20,
                 # verbose = True):

        '''
        Initialize FeatureSelectorRFE instance.

        Parameters
        ----------
        objective : str, optional
            Training objective. Can take values like 'regression', any other value is considered as 'classification'. 
            The default is 'regression'.
        auto : bool, optional
            Verbose flag for short informative messages. The default is True.
            fit_transform has it's own 'verbosity' setting which controlls the 
            verbosity of sklearn.feature_selection.RFECV.
        allowed_score_gap : float, optional
            Auto mode setting: percent difference in CV score on features selected 
            by [RFE with linear model] and by [RFE with RF model]. If the the percent 
            difference in scores (E.g. 0.03) is lower than allowed_score_gap (E.g. 0.05),
            then a smaller final set of features is chosen even if the scoring result is
            marginally lower.
        final_scoring_model : class instance, optional
            Model for cross_val_score of the two sets of features that were indepently
            selected by RFECV with linear model and with RF model. The default is None.
            If None, Lightgbm model will be applied.
        default_model_linear : bool, optional
            Type of default model to use. If True, will deploy LogisticRegression/Ridge, if False will deploy RandomForest.
            The default is False.
        custom_model : class instance, optional
            model class instance that must contain train/predict/predict_proba metods. 
            If not passed, defalult model will be used. The default is None.
        subset_size_mb : int, optional
            Maximum size of data subset for feature selection experiments. 
            Speeds up the feature selection for the large datasets. The default is 20.
        verbose : bool, optional
            Verbose flag for short informative messages. The default is True.
            fit_transform has it's own 'verbosity' setting which controlls the 
            verbosity of sklearn.feature_selection.RFECV.

        Returns
        -------
        None.

        '''       
        self.objective = kwargs.get('objective')
        self.auto = kwargs.get('auto', False)
        self.allowed_score_gap = kwargs.get('allowed_score_gap', 0.0)
        self.final_scoring_model = kwargs.get('final_scoring_model', None)
        self.default_model_linear = kwargs.get('default_model_linear', False)
        self.custom_model = kwargs.get('custom_model', None)
        self.subset_size_mb = kwargs.get('subset_size_mb', 20)
        self.verbose = kwargs.get('verbose', True)
        self._model = self._initialise_model(self.custom_model, self.default_model_linear)
        self.selected_features = None
        self._score_diff_percent = None # placeholder for print results in the console 
        self.printer = Printer(verbose=self.verbose)

    def __repr__(self):
        return f'FeatureSelectorRFE(objective: {self.objective}\
            \n                   model: {self._model}\
            \n                   auto: {self.auto}\
            \n                   allowed_score_gap: {self.allowed_score_gap}\
            \n                   subset_size_mb: {self.subset_size_mb}\
            \n                   verbose: {self.verbose})'
        
    # Validate init arguments
    # =========================================================================
    # define cmmon functions for setters
    def _is_bool(self, val):
        return type(val) == bool
    
    def _is_float(self, val):
        return type(val) == float

    def _is_int(self, val):
        return type(val) == int
    # -------------------------------------------------------------------------
    # objective
    @property
    def objective(self):
        return self._objective

    @objective.setter
    def objective(self, value):
        if type(value) != str : raise TypeError('objective must be a string: "regression" or "classification"')
        self._objective = value
    # -------------------------------------------------------------------------
    # default_model_linear
    @property
    def default_model_linear(self):
        return self._default_model_linear

    @default_model_linear.setter
    def default_model_linear(self, value):
        if not self._is_bool(value) != str : raise TypeError('default_model_linear must be a boolean value: True or False')
        self._default_model_linear = value
    # -------------------------------------------------------------------------
    # custom_model
    @property
    def custom_model(self):
        return self._custom_model
    
    @custom_model.setter
    def custom_model(self, value):
        if value is not None and not hasattr(value, '__dict__'):
            print(f'{value} is not a valid model, it must be a class instance, setting default model')
            self._custom_model = None
        else:
            self._custom_model = value
    # -------------------------------------------------------------------------
    # subset_size_mb
    @property
    def subset_size_mb(self):
        return self._subset_size_mb
    
    @subset_size_mb.setter
    def subset_size_mb(self, value):
        if not self._is_int(value):
            print(f'{value} is not a valid subset_size_mb argument, must be a positive integer, setting to 20')
            self._subset_size_mb = 20
        else:
            if value <= 0:
                print('subset_size_mb must be a positive integer, setting to 20')
                self._subset_size_mb = 20
            else:
                self._subset_size_mb = value
    # -------------------------------------------------------------------------
    # verbose
    @property
    def verbose(self):
        return self._verbose
    
    @verbose.setter
    def verbose(self, value):
        if not self._is_bool(value):
            print(f'{value} is not a valid verbose argument, must be a bool, setting to True')
            self._verbose = True
        else:
            self._verbose = value
    # -------------------------------------------------------------------------
    # auto
    @property
    def auto(self):
        return self._auto
    
    @auto.setter
    def auto(self, value):
        if not self._is_bool(value):
            print(f'{value} is not a valid auto argument, must be a bool, setting to False')
            self._auto = True
        else:
            self._auto = value
    # -------------------------------------------------------------------------
    # final_scoring_model
    @property
    def final_scoring_model(self):
        return self._final_scoring_model
    
    @final_scoring_model.setter
    def final_scoring_model(self, value):
        if value is not None and not hasattr(value, '__dict__'):
            print(f'{value} is not a valid model, it must be a class instance, setting default model, setting to default LGBM')
        else:
            self._final_scoring_model = value
    # -------------------------------------------------------------------------
    # allowed_score_gap
    @property
    def allowed_score_gap(self):
        return self._allowed_score_gap 
    
    @allowed_score_gap.setter
    def allowed_score_gap(self, value):
        if not self._is_float(value):
            print(f'{value} is not a valid auto argument, must be a float, setting to 0.0')
            self._allowed_score_gap = 0.0
        else:
            self._allowed_score_gap = value
    # =========================================================================
    def _initialise_model(self, custom_model, default_model_linear):
        '''Select LogisticRegression/Ridge or RandomForestRegressor/Classifier 
        if no model instance is passed, or use passed model instance'''
        if custom_model is None:
            if default_model_linear:
                return self._get_linear_model()
            else:
                return self._get_randomforest_model()
        else:
            return custom_model
    
    @timer
    def fit_transform(self, X, y, **kwargs):
        '''
        Apply sklearn.feature_selection.RFECV using a model that was defined at init

        Parameters
        ----------
        X : pd.DataFrame
            train features.
        y : pd.Series
            train labels.
        kwargs : optional
            keyword arguments for sklearn.feature_selection.RFECV
            
        Returns
        -------
        pd.DataFrame
            selected train features.

        '''
        self._validate_pandas(X, y)
        self.printer.print('Initiating FeatureSelector', order=1)
        if self.auto:
            self.printer.print(f'Comparing LinearRegression and RandomForest for feature selection', order = 2)
            self._auto_linear_randomforest_selector(X, y, kwargs)
        else:
            self.printer.print(f'Running feature selection with {self._model}', order = 2)
            selector = self._get_selector(self._model, y, kwargs)
            selected_feats_flags = self._prepare_data_apply_selector(X, y, selector, scale_data = self.default_model_linear)
            self.selected_features = X.columns[selected_feats_flags]
        self.printer.print(f'Selected {len(self.selected_features)} features from {X.shape[1]}', order=3)
        return X[self.selected_features]

    def _validate_pandas(self, X, y=None):
        if not isinstance(X, pd.DataFrame):
            raise TypeError('Invalid argument for X, must be a pandas.DataFrame')
        if y is not None:
            if not isinstance(y, pd.Series):
                raise TypeError('Invalid argument y, must be a pandas.Series')

    def _prepare_data_apply_selector(self, X, y, selector, scale_data = False):
        X_subset, y_subset = self._subset_data(X, y)
        if scale_data:
            X_subset = self._scale_data(X_subset)
        try:
            X_subset, y_subset = self._transform_data_to_float_32(X_subset, y_subset)
            selector.fit(X_subset, y_subset)
        except BrokenPipeError:
            print('Experienced BrokenPipeError, restarting')
            selector.fit(X_subset, y_subset)
        feats = selector.support_
        return feats

    def _auto_linear_randomforest_selector(self, X, y, kwargs):
        '''
        Automated feature selection by both built in models LinearRegression & RandomForest.
        
        Both types of models are run independently with RFECV, the resulting sets of features
        are then scored with 2-fold CV by an independent model. Final scoring model can
        be configured at init by passing the model instance to final_scoring_model argument (Lightgbm by default).
        Resulting features are selected taking into account the validation score and 
        allowed_score_gap argument.

        Parameters
        ----------
        X : pd.DataFrame
            train features
        y : pd.Series
            train labels
        kwargs : keyword arguments
            arguments for sklearn.feature_selection.RFECV

        Returns
        -------
        None.

        '''
        from sklearn.model_selection import cross_val_score as cvs

        linear_model = self._get_linear_model()
        randomforest_model = self._get_randomforest_model()
        
        selector_lr = self._get_selector(linear_model, y, kwargs)
        selector_rf = self._get_selector(randomforest_model, y, kwargs)
        
        self.printer.print(f'Running feature selection with {linear_model}', order = 2)
        feats_lr_flags = self._prepare_data_apply_selector(X, y, selector_lr, scale_data = True)

        self.printer.print(f'Running feature selection with {randomforest_model}', order = 2)
        feats_rf_flags = self._prepare_data_apply_selector(X, y, selector_rf, scale_data = False)
        
        scoring = self._get_sklearn_scoring_func_name(y)
        model = self._get_final_scoring_model(y)        
        self.printer.print(f'Scoring selected feats from linear and RF models by: {model}', order = 2)
        score_lr = np.mean(cvs(model, X[X.columns[feats_lr_flags]], y, scoring = scoring, cv = 3))
        score_rf = np.mean(cvs(model, X[X.columns[feats_rf_flags]], y, scoring = scoring, cv = 3))
        self.printer.print(f'RFE by linear model cv-score       : {np.round(score_lr,5)}', order = 4, leading_blank_paragraph=True)
        self.printer.print(f'linear model n selected features   : {len(feats_lr_flags[feats_lr_flags])}', order = 4)
        self.printer.print(f'RFE by RandomForest model cv-score : {np.round(score_rf,5)}', order = 4, leading_blank_paragraph=True)
        self.printer.print(f'RF model n selected features       : {len(feats_rf_flags[feats_rf_flags])}', order = 4)
        
        results_comparison_dict = {
            'linear':{
                'score':score_lr,
                'n_feats':len(feats_lr_flags[feats_lr_flags]),
                'feats_flags':feats_lr_flags
                },
            'RF':{
                'score':score_rf,
                'n_feats':len(feats_rf_flags[feats_rf_flags]),
                'feats_flags':feats_rf_flags
                }
            }      

        selected_model = self._select_based_on_dimensions(results_comparison_dict)
        selected_feats_flags = list(selected_model.values())[0]['feats_flags']
        choice_explanation_message = '' if not self._score_diff_percent else f'\n(validation on these feats is {np.round(self._score_diff_percent,5)}% less accurate but within {self.allowed_score_gap}% allowance)'
        self.printer.print(f'Keeping feats from RFE with {list(selected_model.keys())[0]} model {choice_explanation_message}', order = 3, leading_blank_paragraph=True, breakline='-')
        self.selected_features = X.columns[selected_feats_flags]
    
    def _get_final_scoring_model(self, y):
        '''Set up the default LGBM (regressor or classifier) or use the user defined model'''
        if self.final_scoring_model is None:
            from lightgbm import LGBMClassifier, LGBMRegressor
            if self.objective == 'regression':
                model = LGBMRegressor()
            else:
                if y.nunique() == 2:
                    model = LGBMClassifier(objective = 'binary')
                else:
                    model = LGBMClassifier(objective = 'multiclass')                
        else:
            model = self.final_scoring_model
        return model
        
    # -------------------------------------------------------------------------
    def _select_based_on_dimensions(self, results_comparison_dict):
        '''Header function for features selection from linear & RF feature elimination results'''
        if self.allowed_score_gap == 0:
            selected_model = self._get_result_with_highest_score(results_comparison_dict)
        else:
            selected_model = self._get_results_based_on_allowance(results_comparison_dict)
        return selected_model

    def _get_result_with_highest_score(self, results_comparison_dict):
        '''Return results_comparison_dict subset with the best score'''
        temp_comparison_dict = {}
        for key in results_comparison_dict.keys():
            temp_comparison_dict[key] = results_comparison_dict[key]['score']

        best_score_model_name = max(temp_comparison_dict, key=temp_comparison_dict.get)
        selected_model = {}
        selected_model[best_score_model_name] = results_comparison_dict[best_score_model_name]
        return selected_model

    def _get_results_based_on_allowance(self, results_comparison_dict):
        '''Header function for comparing the scores of both models, number of 
        selected feats and percent difference allowance'''
        selected_model = self._return_model_with_less_feats_and_best_score(results_comparison_dict)
        if len(selected_model.keys()) > 1:
            selected_model = self._extract_model_with_less_feats_if_within_allowance(selected_model)            
        return selected_model            

    def _return_model_with_less_feats_and_best_score(self, results_comparison_dict):
        '''If a model with higher score and less n_feats is yielded return only this model, 
        else return both models'''
        models_names = list(results_comparison_dict.keys())
        temp_comparison_dict = {}
        temp_comparison_dict[models_names[0]] = {}
        temp_comparison_dict[models_names[0]]['smallest_n_feats'] = results_comparison_dict[models_names[0]]['n_feats'] < results_comparison_dict[models_names[1]]['n_feats']
        temp_comparison_dict[models_names[0]]['highest_score'] = results_comparison_dict[models_names[0]]['score'] > results_comparison_dict[models_names[1]]['score']
        temp_comparison_dict[models_names[1]] = {}
        temp_comparison_dict[models_names[1]]['smallest_n_feats'] = results_comparison_dict[models_names[1]]['n_feats'] < results_comparison_dict[models_names[0]]['n_feats']
        temp_comparison_dict[models_names[1]]['highest_score'] = results_comparison_dict[models_names[1]]['score'] > results_comparison_dict[models_names[0]]['score']
        for key, value in temp_comparison_dict.items():
            if np.all(list(value.values())):
                selected_model = {}
                selected_model[key] = results_comparison_dict[key]
                return selected_model
        return results_comparison_dict
        
    def _extract_model_with_less_feats_if_within_allowance(self, selected_model):
        '''Return model with less feats and lower score if score difference percent is within allowance'''
        models_names = list(selected_model.keys())
        temp_comparison_dict = {}
        temp_comparison_dict[models_names[0]] = {}
        temp_comparison_dict[models_names[0]]['lower_score'] = selected_model[models_names[0]]['score'] < selected_model[models_names[1]]['score']
        temp_comparison_dict[models_names[1]] = {}
        temp_comparison_dict[models_names[1]]['lower_score'] = selected_model[models_names[1]]['score'] < selected_model[models_names[0]]['score']
        for key, value in temp_comparison_dict.items():
            if True in value.values():
                lower_score_model = key
            else:
                lower_score_model = self._get_model_with_less_feats_if_scores_are_equal(selected_model)                
        higher_score_model = [key for key in selected_model.keys() if key != lower_score_model][0]
        score_diff_percent = abs(selected_model[lower_score_model]['score'] / selected_model[higher_score_model]['score'] - 1)

        final_model = {}
        if score_diff_percent < self.allowed_score_gap:
            final_model[lower_score_model] = selected_model[lower_score_model]
            self._score_diff_percent = score_diff_percent
        else:
            final_model[higher_score_model] = selected_model[higher_score_model]
        return final_model

    def _get_model_with_less_feats_if_scores_are_equal(self, selected_model):
        '''Return model name with less n_feats if scores are identical'''
        min_n_feats = min(k['n_feats'] for k in selected_model.values()) 
        model_with_min_n_feats = [k for k in selected_model if selected_model[k]['n_feats'] == min_n_feats]
        return model_with_min_n_feats[0]
    
    # -------------------------------------------------------------------------
    def _get_selector(self, model, y, kwargs):
        '''Configure sklearn.feature_selection.RFECV based on target variable and kwargs'''
        final_kwargs = self._set_RFECV_kwargs_from_user_input_and_defaults(kwargs, y)
        selector = RFECV(model, **final_kwargs)
        return selector

    def _set_RFECV_kwargs_from_user_input_and_defaults(self, kwargs, y):
        '''Use user input kwargs for RFECV along with available kwargs for current version of sklearn
        and FeatureSelector default values for kwargs based on sklearn 1.0.1 version to define
        final set of arguments for RFECV. (used to set up RFECV for sklearn 1.0.1 and earlier versions)
        
        '''
        # define defaults for all available kwargs for RFECV in 1.0.1 sklearn version
        sklearn_101_default_kwargs = { 
            'step':1,
            'min_features_to_select':5,
            'cv':2,
            'n_jobs':-1,
            'importance_getter':'auto',
            'scoring':self._get_sklearn_scoring_func_name(y),
            'verbose':0
            }
        available_kwargs = self._get_available_RFECV_params()
        final_kwargs = {}
        for arg in available_kwargs:
            final_kwargs[arg] = kwargs.get(arg, sklearn_101_default_kwargs[arg])
        return final_kwargs

    def _get_available_RFECV_params(self):
        '''Get names of keyword arguments for the user's sklearn.feature_selection.RFECV version'''
        import inspect
        available_args = str(inspect.signature(RFECV)).strip('(').strip(')').split(',')
        available_args = [x for x in available_args if '=' in x]
        available_args = [x.split('=')[0] for x in available_args]
        available_args = [x.replace(' ','') for x in available_args]
        return available_args    
    
    # -------------------------------------------------------------------------
    def _get_linear_model(self):
        '''Return Ridge() or LogisticRegression() from sklearn'''
        if self.objective == 'regression':
            from sklearn.linear_model import Ridge
            return Ridge()
        else:
            from sklearn.linear_model import LogisticRegression
            return LogisticRegression(max_iter=1000)
        
    def _get_randomforest_model(self):
        '''Return RFClassifier or RFRegressor from sklearn'''
        if self.objective == 'regression':
            from sklearn.ensemble import RandomForestRegressor
            return RandomForestRegressor(n_estimators=50, max_depth=2)
        else:
            from sklearn.ensemble import RandomForestClassifier
            return RandomForestClassifier(n_estimators=50, max_depth=2)           
        
    def _subset_data(self, X, y):
        '''Subset train dataset by n megabytes.
        Used to speed up experiments during feature engineering & model tuning.'''
        temp = pd.concat([X, y], axis=1)
        data_size = np.round(temp.memory_usage().sum()/(1024*1024), 2)
    
        if data_size > self.subset_size_mb:
            batch = self.subset_size_mb / data_size
            experimental_data = temp.sample(frac=batch)
            experimental_data.reset_index(drop=True, inplace=True)
            X = experimental_data.drop('target', axis=1)
            y = experimental_data.target
            self.printer.print(f'Data decreased for experiments. Working with {np.round(batch*100,2)}% of data', order = 3)
            del experimental_data
        else:
            self.printer.print('Experiments are carried out on complete dataset', order = 3)
        del temp, data_size
        return X, y  

    def _scale_data(self, X):
        '''Normalize data'''
        from sklearn.preprocessing import StandardScaler
        scaler = StandardScaler()
        X = scaler.fit_transform(X)
        return X

    def _transform_data_to_float_32(self, X_subset, y_subset):
        '''Transform to float32, drop rows with np.nan, np.inf'''
        X_subset = pd.DataFrame(X_subset).astype('float32')
        X_subset = X_subset[~X_subset.isin([np.nan, np.inf, -np.inf]).any(1)]
        y_subset = y_subset.loc[X_subset.index]
        return X_subset, y_subset

    def _get_sklearn_scoring_func_name(self, y):
        '''Return metric name for scoring the feature selection'''
        if self.objective == 'regression':
            return 'neg_root_mean_squared_error'
        else:
            if y.nunique() == 2:
                return 'roc_auc'
            else:
                return 'neg_log_loss'

    def transform(self, test):
        '''Subset data by features that were selected at fit_transform()'''
        self._validate_pandas(test)
        self.printer.print('Applying RFECV (FeatureSelector)', order=1)
        self.printer.print(f'Selected {len(self.selected_features)} features from {test.shape[1]}', order=3)
        return test[self.selected_features]


import pandas as pd
import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers
from tensorflow.keras.layers.experimental import preprocessing
from keras.utils import np_utils
from verstack import Factorizer
tf.get_logger().setLevel('ERROR')

class kerasModel:
    
    def __init__(self, objective, num_layers = 3, epochs = 200, verbose = True):
        '''
        Automatically configurable neural network Regressor/Classifier with mostly fixed architecture.
            - Regressor/Classifier and loss function (mse/binary_crossentropy/categorical_crossentropy are defined by objective
            - number of layers is defined by num_layers argument
            - number of epochs is defined by epochs argument

        Parameters
        ----------
        objective : str
            regression/multiclass/binary for defining the model configuration.
        num_layers : int, optional
            number of layers. can take values from 1 to 3. The default is 3.
        epochs : int, optional
            number of epochs. The default is 200.

        Returns
        -------
        None.

        '''
        self.objective = objective
        self.num_classes = None
        self.num_layers = num_layers
        self.model = None
        self.target_encoder = None
        self.epochs = epochs
        self.verbose = verbose

    def _get_params_from_objective(self, intermediate_activation = False, output_layer = False, loss = False): # self
        '''Define one of the three network parameters
            - intermediate_activation: activation function for all but final layer
            - output_layer: output layer configured based on objective and number of output classes
            - loss: loss function name based on objective
            
        '''
        if intermediate_activation:
            return 'relu' if self.objective == 'regression' else 'tanh'
        if output_layer: 
            if self.objective == 'regression':
                layer = layers.Dense(1)
            elif self.objective == 'binary':
                layer = layers.Dense(2, activation = 'sigmoid')
            else:
                layer = layers.Dense(self.num_classes, activation = 'softmax')
            return layer
        if loss:
            if self.objective == 'regression':
                loss_func_name = 'mean_squared_error'
            elif self.objective == 'binary':
                loss_func_name = 'binary_crossentropy'
            else:
                loss_func_name = 'categorical_crossentropy'
            return loss_func_name

    def _get_architecture(self, normalizer):
        '''Create one of the three predefined architectures with 1/2/3 layers'''
        activation = self._get_params_from_objective(intermediate_activation=True)
        output_layer = self._get_params_from_objective(output_layer=True)
        
        if self.num_layers == 3:
            architecture = [normalizer,
                            layers.Dense(64, activation = activation),
                            layers.Dense(16, activation = activation),
                            layers.Dense(64, activation = activation),
                            output_layer]
        elif self.num_layers == 2:
            architecture = [normalizer,
                            layers.Dense(128, activation = activation),
                            layers.Dense(16, activation = activation),
                            output_layer]
        else:
            architecture = [normalizer,
                            layers.Dense(64, activation = activation),
                            output_layer]
        return architecture

    def build_and_compile_model(self, norm):
        '''Build and compile model'''
        model = keras.Sequential(self._get_architecture(norm))

        model.compile(loss = self._get_params_from_objective(loss=True),
                      optimizer = tf.keras.optimizers.Adam(0.001))
        return model

    def transform_categoric_target(self, y):
        '''Factorize categoric target for classification tasks, save encoder for inverse_transform after predict'''
        enc = Factorizer()
        y = enc.fit_transform(pd.DataFrame(np.array(y), columns = ['target']), 'target')
        y = np_utils.to_categorical(y)
        self.target_encoder = enc
        return y

    def fit(self, X, y):
        if self.verbose:
            verbose = 1
        else:
            verbose = 0
        '''Fit model on X, y, save model to self.model'''
        self.num_classes = y.nunique()
        normalizer = preprocessing.Normalization()
        normalizer.adapt(np.array(X))
        model = self.build_and_compile_model(normalizer)
        if self.objective in ['binary', 'multiclass']:
            y = self.transform_categoric_target(y)
        model.fit(X, y, validation_split=0.2, verbose=verbose, epochs = self.epochs)
        self.model = model
        
    def predict(self, X):
        '''Predict with saved model on X, inverse_transform predictions if classification'''
        pred = self.model.predict(X)          
        if self.objective == 'binary':
            pred = pred[:,1]    
        elif self.objective == 'multiclass':
            pred = np.argmax(pred, 1)
            if self.target_encoder:
                pred = np.array(self.target_encoder.inverse_transform(pd.DataFrame(pred, columns = ['target'])))
        return pred


import unittest
import sys
sys.path.append('../../')
from common import generate_data

from verstack import LGBMTuner

# test overall LGBMTuner not being broken

class TestDateLGBMTuner(unittest.TestCase):
    
    def test_LGBMTuner(self):
        
        df = generate_data(processed=True)
        module = LGBMTuner(metric='rmse', trials=20, visualization=False, refit=True)
        X = df.drop('y', axis = 1)
        y = df['y']
        module.fit(X, y)
        result_trained_model = module.fitted_model is not None
        result_saved_optimized_params = module.best_params is not None
        self.assertTrue(result_trained_model)
        self.assertTrue(result_saved_optimized_params)
        
if __name__ == '__main__':
    unittest.main()


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, usse os.path.abspath to make it absolute, like shown here.
#

import os
import sys
sys.path.insert(0, os.path.abspath('..'))
import guzzle_sphinx_theme
autodoc_mock_imports = ["xgboost", "cmake"]
# -- Project information -----------------------------------------------------

project = 'verstack'
copyright = '2020, Danil Zherebtsov'
author = 'Danil Zherebtsov'

github_doc_root = 'https://github.com/DanilZherebtsov/verstack'

# Version information.
version = '0.1.1'
release = version

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones
extensions = [
#    'matplotlib.sphinxext.plot_directive',
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.mathjax',
    'sphinx.ext.intersphinx',
    #'breathe',
    'recommonmark'
]

graphviz_output_format = 'png'
plot_formats = [('svg', 300), ('png', 100), ('hires.png', 300)]
plot_html_show_source_link = False
plot_html_show_formats = False

# Breathe extension variables

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = ['.rst', '.md']

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

autoclass_content = 'both'

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']
html_extra_path = ['./tmp']

# The reST default role (used for this markup: `text`) to use for all
# documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme_path = guzzle_sphinx_theme.html_theme_path()
html_theme = 'guzzle_sphinx_theme'

# Register the theme as an extension to generate a sitemap.xml
extensions.append("guzzle_sphinx_theme")

# Guzzle theme options (see theme.conf for more information)
html_theme_options = {
    # Set the name of the project to appear in the sidebar
    "project_nav_name": "verstack"
}

html_sidebars = {
  '**': ['logo-text.html', 'localtoc.html']
}

# html_sidebars = { '**': ['globaltoc.html', 'relations.html', 'sourcelink.html', 'searchbox.html'] }

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Output file base name for HTML help builder.
htmlhelp_basename = project + 'doc'

# -- Options for LaTeX output ---------------------------------------------
latex_elements = {
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
  (master_doc, '%s.tex' % project, project,
   author, 'manual'),
]


# hook for doxygen
def run_doxygen(folder):
    """Run the doxygen make command in the designated folder."""
    try:
        retcode = subprocess.call("cd %s; make doxygen" % folder, shell=True)
        if retcode < 0:
            sys.stderr.write("doxygen terminated by signal %s" % (-retcode))
    except OSError as e:
        sys.stderr.write("doxygen execution failed: %s" % e)


def generate_doxygen_xml(app):
    """Run the doxygen make commands if we're on the ReadTheDocs server"""
    read_the_docs_build = os.environ.get('READTHEDOCS', None) == 'True'
    if read_the_docs_build:
        run_doxygen('..')


def setup(app):
    app.add_css_file('custom.css')


import unittest
import sys
sys.path.append('../../')
import numpy as np
from common import generate_data

from verstack import OneHotEncoder

class TestOneHotEncoder(unittest.TestCase):
    
    def test_OneHotEncoder(self):
        df = generate_data()
        # ---------------------
        module = OneHotEncoder()
        module = OneHotEncoder()
        df_train = module.fit_transform(df, 'x')
        result_created_cols = np.all([col in df_train for col in ['a','b','c','d',np.nan]])
        
        self.assertTrue(result_created_cols)
                
if __name__ == '__main__':
    unittest.main()


# classification imports
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.tree import ExtraTreeClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.neighbors import KNeighborsClassifier
from xgboost import XGBClassifier
from lightgbm import LGBMClassifier
from sklearn.svm import SVC
# regression imports
from sklearn.linear_model import Ridge
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.tree import DecisionTreeRegressor
from sklearn.tree import ExtraTreeRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.neighbors import KNeighborsRegressor
from xgboost import XGBRegressor
from lightgbm import LGBMRegressor
from sklearn.svm import SVR
# internal import
from verstack.stacking.kerasModel import kerasModel

def generate_default_layers(objective, epochs = 200, verbose = True):
    '''Initialize two layers with 14 and 2 models

    Parameters
    ----------
    objective : str
        flag to initialize regressors or classifiers.
    epochs : int, optional
        number of epochs for the neural networks compilation. The default is 200.

    Returns
    -------
    layer_1 : list
        14 predefined initialised models.
    layer_2 : list
        2 predefined initialised models.

    '''    
    if objective == 'regression':
        layer_1 = [LGBMRegressor(max_depth = 12, n_jobs = -1),
                   XGBRegressor(max_depth = 10, n_jobs = -1),
                   GradientBoostingRegressor(max_depth = 7),
                   kerasModel(objective = objective, num_layers = 3, epochs = epochs, verbose = verbose),
                   kerasModel(objective = objective, num_layers = 2, epochs = epochs, verbose = verbose),
                   kerasModel(objective = objective, num_layers = 1, epochs = epochs, verbose = verbose),
                   ExtraTreeRegressor(max_depth = 12),
                   RandomForestRegressor(max_depth = 7, n_jobs = -1),
                   LinearRegression(),
                   KNeighborsRegressor(n_neighbors=15, n_jobs = -1),
                   KNeighborsRegressor(n_neighbors=10, n_jobs = -1),
                   SVR(kernel = 'rbf'),
                   DecisionTreeRegressor(max_depth = 15), 
                   DecisionTreeRegressor(max_depth = 8)]

        layer_2 = [LGBMRegressor(max_depth = 3), 
                   Ridge()]
    else:
        layer_1 = [LGBMClassifier(max_depth = 12, n_jobs = -1),
                   XGBClassifier(max_depth = 10, n_jobs = -1),
                   GradientBoostingClassifier(max_depth = 7),
                   kerasModel(objective = objective, num_layers = 3, epochs = epochs, verbose = verbose),
                   kerasModel(objective = objective, num_layers = 2, epochs = epochs, verbose = verbose),
                   kerasModel(objective = objective, num_layers = 1, epochs = epochs, verbose = verbose),
                   ExtraTreeClassifier(max_depth = 12),
                   RandomForestClassifier(max_depth = 7, n_jobs = -1),
                   LogisticRegression(),
                   KNeighborsClassifier(n_neighbors=15, n_jobs = -1),
                   KNeighborsClassifier(n_neighbors=10, n_jobs = -1),
                   SVC(kernel = 'rbf'),
                   DecisionTreeClassifier(max_depth = 15),
                   DecisionTreeClassifier(max_depth = 8)]

        layer_2 = [LGBMClassifier(max_depth = 3), 
                   LogisticRegression()]   

    return layer_1, layer_2




import numpy as np
import pandas as pd
import random

binary_target_choices = [0,1]
multiclass_target_choices = [0,1,2,3]
cat_col_choices = ['a', 'b', 'c', 'd']
date_col_choices = [
    '2015-06-17',
    '2015-05-31',
    '2015-08-28',
    '2015-06-29',
    '2015-03-17',
    '2015-02-25',
    '2015-01-12',
    '2015-11-29',
    '2015-12-29',
    '2015-12-12',
    '2015-12-29',
    '2015-09-30',
    '2015-07-01',
    '2015-06-30',
    '2015-05-30',
    '2015-04-23',
    '2015-07-25',
    '2015-04-30',
    '2015-07-31'
    ]

def generate_data(processed=False):
    # generate data either unprocessed or processed (ready for modeling)
    np.random.seed(42)
    random.seed(42)
    if not processed:
        var_x = [random.choice(cat_col_choices) for i in range(100)]
        var_timestamp = [random.choice(date_col_choices) for i in range(100)]
        numeric_target = list(np.random.uniform(0,1,100))
        binary_target = [random.choice(binary_target_choices) for i in range(100)]
        df = pd.DataFrame({'x':var_x, 'y':numeric_target, 'y_binary':binary_target, 'timestamp':var_timestamp})
        df.loc[2,'x'] = np.nan
        df.loc[3,'timestamp'] = np.nan
    else:
        var_x = list(np.random.uniform(0,1,100))
        var_z = list(np.random.uniform(0,1,100))
        numeric_target = list(np.random.uniform(0,1,100))
        binary_target = [random.choice(binary_target_choices) for i in range(100)]
        multiclass_target = [random.choice(multiclass_target_choices) for i in range(100)]
        df = pd.DataFrame({'x':var_x, 'z':var_z, 'y':numeric_target, 'y_binary':binary_target, 'y_multiclass':multiclass_target})
    return df




import unittest
import sys
sys.path.append('../../')
from common import generate_data

from verstack import MeanTargetEncoder

class TestMeanTargetEncoder(unittest.TestCase):
    
    def test_MeanTargetEncoder(self):
        df = generate_data()
        # ---------------------
        module = MeanTargetEncoder()
        df_train = module.fit_transform(df, 'x', 'y')
        df_test = module.transform(df)
        inverse_df_test = module.inverse_transform(df_test)

        result_train_transform = df_train['x'].dtype==float
        result_test_transform = df_test['x'].dtype==float
        result_test_inverse_transform = inverse_df_test['x'].dtype=='O'

        self.assertTrue(result_train_transform)
        self.assertTrue(result_test_transform)
        self.assertTrue(result_test_inverse_transform)
        
if __name__ == '__main__':
    unittest.main()


import unittest
import sys
sys.path.append('../../')
from common import generate_data

from verstack import DateParser

# test overall DateParser not being broken

class TestDateParser(unittest.TestCase):
    
    def test_DateParser(self):
        df = generate_data()
        module = DateParser(country='Russia', payday=[15,30])    
        transformed = module.fit_transform(df)
        result = len(module.datetime_cols)==1
        self.assertTrue(result)
        
if __name__ == '__main__':
    unittest.main()


import unittest
import sys
sys.path.append('../../')
import numpy as np
from common import generate_data

from verstack import Factorizer

class TestFactorizer(unittest.TestCase):
    
    def test_Factorizer(self):
        df = generate_data()
        # ---------------------
        module = Factorizer()
        df_train = module.fit_transform(df, 'x')
        df_test = module.transform(df)

        inverse_df_train = module.inverse_transform(df_train)
        inverse_df_test = module.inverse_transform(df_test)

        result_transform = np.all(df_train['x'] == df_test['x'])
        result_inverse_transform = np.all(inverse_df_train['x'].dropna() == df['x'].dropna())

        self.assertTrue(result_transform)
        self.assertTrue(result_inverse_transform)
        
if __name__ == '__main__':
    unittest.main()


import operator
import numpy as np
import pandas as pd
import os
import lightgbm as lgb
import optuna
from sklearn.model_selection import train_test_split
from copy import copy
from functools import partial
import warnings
from verstack.tools import timer, Printer
warnings.filterwarnings("ignore")
from verstack.lgbm_optuna_tuning.lgb_metrics import get_pruning_metric, get_optimization_metric_func, define_objective, classification_metrics, regression_metrics, get_eval_score, print_lower_greater_better, supported_metrics, get_n_rounds_optimization_metric
from verstack.lgbm_optuna_tuning.args_validators import *
from verstack.lgbm_optuna_tuning.optuna_tools import Distribution, OPTUNA_DISTRIBUTIONS_MAP, SearchSpace
# BACKLOG: add option to pass different init params on class __init__
# BACKLOG: add option to pass different param_grid for optimization
# OPTIONAL TODO - add factorization to targets and inverse_transform predictions for classification (E.g. Ghosts, prudential)

class LGBMTuner:

    __version__ = '0.0.12'

    def __init__(self, metric, trials = 100, refit = True, verbosity = 1, visualization = True, seed = 42, eval_results_callback = None):
        '''
        Class to automatically tune LGBM model with optuna.
        
        Model type (regressor/classifier) is inferred based on target variable & metric.
        Init parameters and search space are inferred by built in logic.

        Parameters
        ----------
        metric : str
            evaluation metric name.
        trials : int, optional
            number of hyperparameter search trials. The default is 200.
        refit : bool, optional
            Flag to refit the model with optimized parameters on whole dataset. 
            The default is True.
        verbosity : int, optional
            console verbosity level: 0 - no output except for optuna CRITICAL errors and builtin exceptions; 
            (1-5) based on optuna.logging options. The default is 1.
        visualization : bool, optional
            flag to print optimization & feature importance plots. The default is True.
        seed : int
            random_state.
        eval_results_callback : func
            callback function to be applied on the eval_results dictionary that is being populated
            with evaluation metric score upon completion of each training trial.

        Returns
        -------
        None.

        '''
        self.verbosity = verbosity
        self.printer = Printer(verbose=True if self.verbosity > 0 else False)
        self.metric = metric
        self.trials = trials
        self.refit = refit        
        self.visualization = visualization
        self.seed = seed
        self.target_minimum = None
        self._fitted_model = None
        self._feature_importances = None
        self._study = None # save optuna study for plotting
        self.target_classes = None
        self._init_params = None
        self._best_params = None
        self.eval_results = {} # evaluation metric results per each trial storage
        self.eval_results_callback = eval_results_callback

    # print init parameters when calling the class instance
    def __repr__(self):
        return f'LGBMTuner(Evaluation metric: {self._metric}\
            \n          trials: {self.trials}\
            \n          refit: {self.refit}\
            \n          verbosity: {self.verbosity}\
            \n          visualization: {self.visualization})'

    # Validate init arguments
    # =========================================================================
    # define cmmon functions for setters
    def _is_bool(self, val):
        return type(val) == bool

    def _is_int(self, val):
        return type(val) == int
    # -------------------------------------------------------------------------
    # metric
    @property
    def metric(self):
        return self._metric

    @metric.setter
    def metric(self, value):
        if type(value) != str : raise TypeError('metric must be a string')
        if value not in supported_metrics: raise KeyError(f'LGBMTuner supports the following evaluation metrics: {supported_metrics}')
        self._metric = value
    # -------------------------------------------------------------------------
    # trials
    @property
    def trials(self):
        return self._trials

    @trials.setter
    def trials(self, value):
        if not self._is_int(value) : raise TypeError('trials must be an integer')
        if value <= 0: raise ValueError('trials must be an integer > 0')
        self._trials = value
    # -------------------------------------------------------------------------
    # refit
    @property
    def refit(self):
        return self._refit

    @refit.setter
    def refit(self, value):
        if not self._is_bool(value) : raise TypeError('acceptable refit options are True/False')
        self._refit = value
    # -------------------------------------------------------------------------
    # verbosity
    @property
    def verbosity(self):
        return self._verbosity

    @verbosity.setter
    def verbosity(self, value):
        if not self._is_int(value) : raise TypeError('verbosity must be an integer')
        if value not in range(0,6) : raise ValueError('acceptable verbosity options are [0,1,2,3,4,5]')
        self._verbosity = value
    # -------------------------------------------------------------------------
    # visualization
    @property
    def visualization(self):
        return self._visualization

    @visualization.setter
    def visualization(self, value):
        if not self._is_bool(value) : raise TypeError('acceptable visualization options are True/False')
        self._visualization = value
    # -------------------------------------------------------------------------    
    # seed
    @property
    def seed(self):
        return self._seed

    @seed.setter
    def seed(self, value):
        if not self._is_int(value) : raise TypeError('seed must be an integer')
        self._seed = value
    # -------------------------------------------------------------------------    
    # only getters for the following arguments
    # -------------------------------------------------------------------------    
    # init_params
    @property
    def init_params(self):
        return self._init_params
    # -------------------------------------------------------------------------    
    # best_params
    @property
    def best_params(self):
        return self._best_params
    # -------------------------------------------------------------------------    
    # feature_importances
    @property
    def feature_importances(self):
        if not self._fitted_model:
            raise AttributeError('LGBMTuner.fit(refit = True) must be applied before feature_importances can be displayed')
        return self._feature_importances
    # -------------------------------------------------------------------------    
    # fitted_model
    @property
    def fitted_model(self):
        return self._fitted_model
    # -------------------------------------------------------------------------    
    # study
    @property
    def study(self):
        return self._study
    # -------------------------------------------------------------------------
    # eval_results_callback
    @property
    def eval_results_callback(self):
        return self._eval_results_callback

    @eval_results_callback.setter
    def eval_results_callback(self, value):
        if not value is None:
            if not hasattr(value, '__call__') : raise TypeError('eval_results_callback must be a function')
        self._eval_results_callback = value
    # -------------------------------------------------------------------------    
    def _init_params_on_input(self, rows_num, y):
        '''
        Get model parameters depending on dataset parameters.

        Parameters
        ----------
        rows_num : int
            number of rows in training data.
        y : pd.Series
            train labels

        Returns
        -------
        None.

        '''
        # TODO: use number of features
        
        # default lgbm parameters - basis for minor changes based on dataset length
        default_params_classification = {"task": "train",
                                        "learning_rate": 0.05,
                                        "num_leaves": 128,
                                        "colsample_bytree": 0.7,
                                        "subsample": 0.7,
                                        "bagging_freq": 1,
                                        "max_depth": -1,
                                        "verbosity": -1,
                                        "reg_alpha": 1,
                                        "reg_lambda": 0.0,
                                        "min_split_gain": 0.0,
                                        "zero_as_missing": False,
                                        "max_bin": 255,
                                        "min_data_in_bin": 3,
                                        "n_estimators": 3000,
                                        "early_stopping_rounds": 100,
                                        "random_state": 42}

        default_params_regression = {"learning_rate": 0.05,
                                    "num_leaves": 32,
                                    "colsample_bytree": 0.9,
                                    "subsample": 0.9,
                                    "verbosity": -1,
                                    "n_estimators": 3000,
                                    "early_stopping_rounds": 100,
                                    "random_state": 42}


        task = define_objective(self.metric, y)
        
        if task == 'regression':
            self._init_params = copy(default_params_regression)
        else:
            self._init_params = copy(default_params_classification)

        # define additional params based on task type
        if task == 'binary':
            self._init_params['num_classes'] = 1 
        elif task == 'multiclass':
            self._init_params['num_classes'] = y.nunique()
        # -------------------------------------------------------------------------
        # populate objective and metric based
        self._init_params['objective'] = define_objective(self.metric, self.target_classes)
        self._init_params['metric'] = get_pruning_metric(self.metric, self.target_classes)
        # -------------------------------------------------------------------------
        # do not use all available threads and make sure cpu_count is not negative
        cpu_count = int(np.where(os.cpu_count()-2 < 0, 1, os.cpu_count()-2))

        self._init_params['num_threads'] = cpu_count

        if rows_num <= 10000:
            init_lr = 0.01
            ntrees = 3000
            es = 200
        elif rows_num <= 20000:
            init_lr = 0.02
            ntrees = 3000
            es = 200
        elif rows_num <= 100000:
            init_lr = 0.03
            ntrees = 1200
            es = 200
        elif rows_num <= 300000:
            init_lr = 0.04
            ntrees = 2000
            es = 100
        else:
            init_lr = 0.05
            ntrees = 2000
            es = 100

        if rows_num > 300000:
            self._init_params["num_leaves"] = 128 if task == "regression" else 244
        elif rows_num > 100000:
            self._init_params["num_leaves"] = 64 if task == "regression" else 128
        elif rows_num > 50000:
            self._init_params["num_leaves"] = 32 if task == "regression" else 64
            # params['reg_alpha'] = 1 if task == 'reg' else 0.5
        elif rows_num > 20000:
            self._init_params["num_leaves"] = 32 if task == "regression" else 32
            self._init_params["reg_alpha"] = 0.5 if task == "regression" else 0.0
        elif rows_num > 10000:
            self._init_params["num_leaves"] = 32 if task == "regression" else 64
            self._init_params["reg_alpha"] = 0.5 if task == "regression" else 0.2
        elif rows_num > 5000:
            self._init_params["num_leaves"] = 24 if task == "regression" else 32
            self._init_params["reg_alpha"] = 0.5 if task == "regression" else 0.5
        else:
            self._init_params["num_leaves"] = 16 if task == "regression" else 16
            self._init_params["reg_alpha"] = 1 if task == "regression" else 1

        self._init_params["learning_rate"] = init_lr
        self._init_params["n_estimators"] = ntrees
        self._init_params["early_stopping_rounds"] = es
    # -----------------------------------------------------------------------------

    def _get_default_search_space(self, estimated_n_trials):
        '''Sample hyperparameters from suggested.

        Parameters
        ----------
        estimated_n_trials : int
            number of optuna trails.

        Returns
        -------
        search_space : dict
            optimization search space grid and type of distribution for each parameter.

        '''
        # TODO: create addditional options based on bigger estimated_n_trials
        search_space = {}

        search_space["colsample_bytree"] = SearchSpace(Distribution.UNIFORM, low=0.5, high=1.0) 
        search_space["num_leaves"] = SearchSpace(Distribution.INTUNIFORM, low=16, high=255)

        if estimated_n_trials > 30:
            search_space["subsample"] = SearchSpace(Distribution.UNIFORM, low=0.5, high=1.0) 
            search_space["min_sum_hessian_in_leaf"] = SearchSpace(Distribution.LOGUNIFORM, low=1e-3, high=10.0)

        if estimated_n_trials > 100:
            search_space["reg_alpha"] = SearchSpace(Distribution.LOGUNIFORM, low=1e-8, high=10.0)
            search_space["reg_lambda"] = SearchSpace(Distribution.LOGUNIFORM, low=1e-8, high=10.0)

        return search_space
    # -----------------------------------------------------------------------------

    def _sample_from_search_space(self, optimization_search_space, trial):
        '''
        Get params for a trial.

        Parameters
        ----------
        optimization_search_space : dict
            grid for selecting parameters.
        trial : optuna.Trial
            trial object.
        suggested_params : dict
            init params on input.

        Raises
        ------
        ValueError
            Optuna distribution error.

        Returns
        -------
        trial_values : dict
            trial parameters consisting from suggested_params modified by optimization_search_space.

        '''
        trial_values = copy(self._init_params)
        for parameter, SearchSpace in optimization_search_space.items():
            if SearchSpace.distribution_type in OPTUNA_DISTRIBUTIONS_MAP:
                trial_values[parameter] = getattr(trial, OPTUNA_DISTRIBUTIONS_MAP[SearchSpace.distribution_type])(
                    name=parameter, **SearchSpace.params)
            else:
                for key, value in optimization_search_space.items():
                    print(key)
                    print(value.distribution_type)
                raise ValueError(f"Optuna does not support distribution {SearchSpace.distribution_type}")
        return trial_values
    # -----------------------------------------------------------------------------

    def _get_dtrain_dvalid_objects(self, X, y, metric, seed = None, return_raw_valid = False):
        '''
        Create lgbm.Datasets for training and validation.
        
        By default splits without defined random_state in order to replicate CV (sort of).
        Seed is used for optimize_n_estimators.

        Parameters
        ----------
        X : pd.DataFrame
            train features.
        y : pd.Series
            train labels.
        metric : str
            evaluation metric name.
        seed : int/None
            random state for split
        return_raw_valid : bool
            Flag to return valid_x, valid_y in addition to dvalid (for get_validation_score())

        Returns
        -------
        dtrain : lgbm.Dataset
            train data.
        dvalid : ldgb.Dataset
            valid data.

        '''
        if metric in classification_metrics:
            train_x, valid_x, train_y, valid_y = train_test_split(X, y, test_size=0.25, stratify = y)
        else:
            train_x, valid_x, train_y, valid_y = train_test_split(X, y, random_state = seed, test_size=0.25)
        dtrain = lgb.Dataset(train_x, label=train_y)
        dvalid = lgb.Dataset(valid_x, label=valid_y)
        if return_raw_valid:
            return dtrain, dvalid, valid_x, valid_y
        else:
            return dtrain, dvalid
    # -----------------------------------------------------------------------------

    def optimize_n_estimators(self, X, y, params, verbose_eval = 100):
        '''
        Optimize n_estimators for lgb model.
        
        Here (after all the tuning) the actual eval_metric is used for early stopping.
        get_n_rounds_optimization_metric(self.metric) returns the feval function 
        from lgb_metrics in an acceptable format.

        Parameters
        ----------
        X : np.array
            train features.
        y : np.array
            train target.
        params : dict
            model parameters.

        Returns
        -------
        int
            best_iteration for further n_estimators param.
        best_score : float
            validation score from best_iteration

        '''
        validate_numpy_ndarray_arguments(X)
        validate_numpy_ndarray_arguments(y)
        validate_params_argument(params)
        validate_verbose_eval_argument(verbose_eval)

        self.printer.print('Tune n_estimators with early_stopping', order=2)
        if self.verbosity>0:
            verbose_eval_rounds = verbose_eval
        else:
            verbose_eval_rounds = None

        # temporarily change the params['metric'] for the best performance on the eval_metric
        n_rounds_optimization_params = params.copy()
        n_rounds_optimization_params['metric'] = 'None'

        dtrain, dvalid = self._get_dtrain_dvalid_objects(X, y, self.metric, seed = self.seed)

        if n_rounds_optimization_params['objective'] == 'multiclass':
            # disable custom feval function for multiclass 
            # lgbm predicts multiple classes in a single array 
            # without a way to assign predicted probability to a certain class
            # use built in multi_logloss instead
            del n_rounds_optimization_params['metric']
            lgb_model = lgb.train(n_rounds_optimization_params, 
                                  dtrain, 
                                  10000, 
                                  valid_sets=[dtrain, dvalid], 
                                  valid_names=['train', 'valid'],
                                  verbose_eval = verbose_eval_rounds, 
                                  early_stopping_rounds=self._init_params['early_stopping_rounds'])
        else:
            lgb_model = lgb.train(n_rounds_optimization_params, 
                                  dtrain, 
                                  10000, 
                                  valid_sets=[dtrain, dvalid], 
                                  valid_names=['train', 'valid'],
                                  verbose_eval = verbose_eval_rounds, 
                                  early_stopping_rounds=self._init_params['early_stopping_rounds'],
                                  feval = get_n_rounds_optimization_metric(self.metric))
        
        best_score = list(lgb_model.best_score['valid'].values())[0]
        
        return lgb_model.best_iteration, best_score

    # ------------------------------------------------------------------------------------------
    def fit_optimized(self, X, y):
        '''
        Train model with tuned params on whole train data

        Parameters
        ----------
        X : np.array
            train features.
        y : np.array
            train target.

        Returns
        -------
        None.

        '''
        validate_numpy_ndarray_arguments(X)
        validate_numpy_ndarray_arguments(y)
        
        import lightgbm as lgb
        self.printer.print('Fitting optimized model with the follwing params:', order=2)
        for key, value in self._best_params.items():
            print(f'{key:<33}: {value}')
        self._fitted_model = lgb.train(self._best_params, lgb.Dataset(X,y))

    # ------------------------------------------------------------------------------------------
    def _get_target_minimum(self, y):
        '''Record target minimum value for replacing negative predictions in regression.'''
        if self.metric in regression_metrics:    
            self.target_minimum = min(y)

# ------------------------------------------------------------------------------------------

    def _get_validation_score(self, trial, dtrain, dvalid, valid_x, valid_y, optimization_metric_func, params, pruning_callback):
        '''
        Train model with trial params and validate on valid set against the defined metric.
        Print optimization result every iteration.
        If evaluation metric != optimization metric, print evaluation metric.

        Parameters
        ----------
        trial : optuna.Trial #1
            parameters tuning iteration.
        dtrain : lgb.Dataset 
            train data.
        dvalid : lgb.Dataset 
            valid data.
        valid_x : pd.DataFrame 
            valid features.
        valid_y : pd.Series 
            valid target.
        optimization_metric_func : func 
            scorer function.
        params : dict #1
            model parameters.
        pruning_callback : func
            callback function.

        Returns
        -------
        result : float
            validation result.

        '''
        gbm = lgb.train(params, dtrain, valid_sets=[dvalid], callbacks=[pruning_callback])
        pred = gbm.predict(valid_x)

        result = optimization_metric_func(valid_y, pred)

        optimization_direction = 'lower-better'

        self.printer.print(f'Trial number: {trial.number} finished', order=3)
        self.printer.print(f'Optimization score ({optimization_direction:<4}): {optimization_metric_func.__name__}: {result}', order=4)
        # save evaluation metric results per each trial
        self.eval_results[f'train_trial_{trial.number}'] = result
        # calculate & print eval_metric only if eval_metric != optimization_metric
        if self.metric != optimization_metric_func.__name__:
            eval_score = get_eval_score(valid_y, pred, self.metric, params['objective'])
            self.printer.print(f'Evaluation score ({print_lower_greater_better(self.metric):<4}): {self.metric}: {eval_score}', order=4)
            # save evaluation metric results per each trial
            self.eval_results[f'train_trial_{trial.number}'] = eval_score
        self.printer.print(breakline='.')

        if self.eval_results_callback:
            self.eval_results_callback(self.eval_results)

        return result

    # ------------------------------------------------------------------------------------------
    def _objective(self, trial, X, y):
        '''
        Define objective for optuna tiral training.
        
        - Create optimization metric based on evaluatin metric passed by user
        - Create train/valid splits
        - Define suggested initial parameters for LGB model based on data and eveluation metric
        - Define the search space for some params within suggested parameters
        - Create final params dict based on suggested initial params and the a step in optimization search space grid
        - Define a pruning callback to stop training of unpromissed trails
        - Train a given trail & validate by optimization_metric_func

        Parameters
        ----------
        trial : optuna.Trial
            training step.
        X : pd.DataFrame
            train features.
        y : pd.Series
            train labels.
        metric : str
            evaluation metric name.

        Returns
        -------
        result : float
            optimization metric validation result.

        '''
        optimization_metric_func = get_optimization_metric_func(self.metric)
        dtrain, dvalid, valid_x, valid_y = self._get_dtrain_dvalid_objects(X, y, self.metric, return_raw_valid = True)
        optimization_search_space = self._get_default_search_space(estimated_n_trials = 200)    
        params = self._sample_from_search_space(optimization_search_space, trial)
        # Add a callback for pruning.
        pruning_callback = optuna.integration.LightGBMPruningCallback(trial, get_pruning_metric(self.metric, self.target_classes))
        result = self._get_validation_score(trial, dtrain, dvalid, valid_x, valid_y, optimization_metric_func, params, pruning_callback)

        return result 
    # ------------------------------------------------------------------------------------------
    def _check_refit_status(self, method):
        if not self.refit:
            raise ModuleNotFoundError(f'LGBMTuner.fit(refit = True) must be applied before {method} execution')
    # ------------------------------------------------------------------------------------------
    
    def predict(self, test, threshold = 0.5):
        '''
        Predict on test set with fitted model.
        
        For regression replace negative preds by self.target_minimum (if train labels
        were not negative).

        Parameters
        ----------
        test : pd.DataFrame
            test features.
        threshold : float, optional
            binary classification probability threshold. The default is 0.5.

        Returns
        -------
        pred : array
            predicted values.

        '''    
        self._check_refit_status('predict()')
        validate_features_argument(test)
        validate_threshold_argument(threshold)

        pred = self._fitted_model.predict(test.values, num_threads = self._init_params['num_threads'])
        
        # regression predict
        if self.metric in regression_metrics:
            if self.target_minimum >= 0:
                pred = np.where(pred < 0, self.target_minimum, pred)
            return pred
        else:
            # binary classification predict
            if self._fitted_model.params['objective'] == 'binary':
                pred = (pred > threshold).astype('int')
            # multiclass classification predict
            # TODO: SAVE CLASSES NAMES TO APPLY THEM BACK
            else:
                pred = np.argmax(pred, axis = 1)
            return pred
    # ------------------------------------------------------------------------------------------
    
    def predict_proba(self, test):
        '''
        Predict on test set (classification only) and return probabilities.

        Parameters
        ----------
        test : pd.DataFrame
            test features.

        Raises
        ------
        TypeError
            If self._fitted_model.params['objective'] == 'regression - notify
            that predict_proba() is for classification objectives only.

        Returns
        -------
        pred : array
            predicted values.

        '''
        self._check_refit_status('predict_proba()')
        validate_features_argument(test)

        if self.metric in regression_metrics:
            raise TypeError('predict_proba() is applicable for classification problems only')
        pred = self._fitted_model.predict(test.values, num_threads = self._init_params['num_threads'])
        return pred
    # ------------------------------------------------------------------------------------------
    
    def _populate_best_params_to_init_params(self, best_params):
        '''Populate the learned params into the suggested params'''
        # output params are temporary, because n_estimators tuning will follow
        temp_params = copy(self._init_params)
        for key, val in best_params.items():
            temp_params[key] = val
        # remove early_stopping & n_estimators from params (used during optuna optimization).
        # final early stopping will be trained during final_estimators_tuning
        del temp_params['early_stopping_rounds']
        del temp_params['n_estimators']
        
        return temp_params
    # ------------------------------------------------------------------------------------------

    def _save_feature_importances(self, train_features):
        '''Save feature importances in class instance as a pd.Series'''
        feat_importances = pd.Series(self._fitted_model.feature_importance(), index = train_features)
        normalized_importances = np.round((lambda x: x/sum(x))(feat_importances),5)
        self._feature_importances = normalized_importances
    # ------------------------------------------------------------------------------------------

    def _display_html(self, html_file):
        '''Run html plot in the default browser'''
        import webbrowser
        import time
        # 1st method how to open html files in chrome using
        filename = 'file:///'+os.getcwd()+'/' + html_file
        webbrowser.open_new_tab(filename)
        time.sleep(2)
        os.remove(html_file)
    # ------------------------------------------------------------------------------------------

    def plot_importances(self, n_features=15, figsize=(10,6), interactive=False, display=True, plotly_fig_update_layout_kwargs={}):
        '''
        Plot feature importances.

        Parameters
        ----------
        n_features : int, optional
            Number of features to plot. The default is 15.
        figsize : tuple, optional
            Figure size. The default is (10,6).
        interactive : bool, optional
            Create & save to current wd interactive html plot. The default is False.
        display: bool, optional
            Display plot in browser. If False, plot will be saved in cwd. The default is True.
        plotly_fig_update_layout_kwargs: dict, optional
            kwargs for plotly.fig.update_layout() function. The default is empty dict and default_plotly_fig_update_layout_kwargs will be used

        Returns
        -------
        None.

        '''
        self._check_refit_status('plot_importances()')
        validate_plot_importances_n_features_argument(n_features)
        validate_plot_importances_figsize_argument(figsize)
        validate_plotting_interactive_argument(interactive)

        if interactive:
            import plotly.express as px
            plot_df = pd.DataFrame(self._feature_importances.nlargest(n_features).sort_values(), columns = ['importance'])
            plot_df['feature'] = plot_df.index
            plot_df.reset_index(drop = True, inplace = True)
            plot_df['importance_percent'] = [str(np.round(val*100,3))+'%' for val in plot_df['importance']]
            
            BAR_COLOR = '#F99245'
            
            fig = px.bar(plot_df, 
                         x = 'importance', 
                         y = 'feature', 
                         orientation = 'h',
                         color_discrete_sequence = [BAR_COLOR]*len(plot_df),
                         text = 'importance_percent')

            default_plotly_fig_update_layout_kwargs = {'plot_bgcolor':'#20253c', # plot color
                                                       'paper_bgcolor':'#2d3250', # html color
                                                       'font_color':'white',
                                                       'title_font_color':'lightgrey',
                                                       'xaxis':{'visible':False,
                                                                'showticklabels':False,
                                                                'showgrid':False},
                                                       'yaxis':{'showticklabels':True,
                                                                'title':''}}

            # set plotly update_layout kwargs            
            if plotly_fig_update_layout_kwargs:
                kwargs = plotly_fig_update_layout_kwargs
            else:
                kwargs = default_plotly_fig_update_layout_kwargs

            fig.update_layout(**kwargs)

            fig.update_traces(textposition='inside',
                              marker_line_color=BAR_COLOR)

            fig.write_html('feature_importance_plot.html', config={'displaylogo': False})
            if display:
                try:
                    self._display_html('feature_importance_plot.html')
                except Exception as e:
                    print(f'Display html error: {e}')
                    print(f'Optimization History Plot is saved to {os.path.join(os.getcwd(), "feature_importance_plot.html")}')
        else:
            import matplotlib.pyplot as plt
            importances_for_plot = self._feature_importances.nlargest(n_features).sort_values()
            importances_for_plot.plot(kind = 'barh', figsize = figsize, color = 'grey')
            for i, v in enumerate(importances_for_plot):
                plt.text(v, i, str(v), color='grey', fontsize = 10, va  = 'center')
            plt.title('Feature importances (sum up to 1)', color = 'grey')
            plt.show()
    # ------------------------------------------------------------------------------------------
    
    def plot_optimization_history(self, interactive=False, display=True):
        '''
        Plot parameters optimization history.

        Parameters
        ----------
        interactive : bool, optional
            Create & save to current wd interactive html plot. The default is False.
        display: bool, optional
            Display plot in browser. If False, plot will be saved in cwd. The default is True.

        Returns
        -------
        None.

        '''
        self._check_refit_status('plot_optimization_history()')
        validate_plotting_interactive_argument(interactive)
        
        if interactive:
            from optuna.visualization import plot_optimization_history
            fig = plot_optimization_history(self._study)
            fig.write_html("optimization_history_plot.html")
            if display:
                try:
                    self._display_html("optimization_history_plot.html")
                except Exception as e:
                    print(f'Display html error: {e}')
                    print(f'Optimization History Plot is saved to {os.path.join(os.getcwd(), "optimization_history_plot.html")}')
        else:
            from optuna.visualization.matplotlib import plot_optimization_history
            import matplotlib.pyplot as plt
            plot_optimization_history(self._study)
            plt.show()
    # ------------------------------------------------------------------------------------------
    
    def plot_param_importances(self, interactive=False, display=True):
        '''
        Plot parameters importance.

        Parameters
        ----------
        interactive : bool, optional
            Create & save to current wd interactive html plot. The default is False.
        display: bool, optional
            Display plot in browser. If False, plot will be saved in cwd. The default is True.

        Returns
        -------
        None.

        '''
        self._check_refit_status('plot_param_importance()')
        validate_plotting_interactive_argument(interactive)

        if interactive:
            from optuna.visualization import plot_param_importances
            fig = plot_param_importances(self._study)
            fig.write_html("param_importances_plot.html")
            if display:
                try:
                    self._display_html('param_importances_plot.html')
                except Exception as e:
                    print(f'Display html error: {e}')
                    print(f'Param Importances Plot is saved to {os.path.join(os.getcwd(), "param_importances_plot.html")}')
        else:
            from optuna.visualization.matplotlib import plot_param_importances
            import matplotlib.pyplot as plt
            plot_param_importances(self._study)
            plt.show()
    # ------------------------------------------------------------------------------------------
    
    def plot_intermediate_values(self, interactive=False, legend=False, display=True):
        '''
        Plot optimization trials history. Shows successful and terminated trials.

        Parameters
        ----------
        interactive : bool, optional
            Create & show in default browsersave to current wd interactive html plot. The default is False.
        legend : bool, optional
            Flag to include legend in the static (not interactive) plot. The default is False.
        display: bool, optional
            Display plot in browser. If False, plot will be saved in cwd. The default is True.

        Returns
        -------
        None.

        '''        
        self._check_refit_status('plot_intermediate_values()')
        validate_plotting_interactive_argument(interactive)
        validate_plotting_legend_argument(legend)
        
        if interactive:
            from optuna.visualization import plot_intermediate_values
            fig = plot_intermediate_values(self._study)
            fig.write_html("intermediate_values_plot.html")
            if display:
                try:
                    self._display_html('intermediate_values_plot.html')
                except Exception as e:
                    print(f'Display html error: {e}')
                    print(f'Intermediate Values Plot is saved to {os.path.join(os.getcwd(), "intermediate_values_plot.html")}')
        else:
            from optuna.visualization.matplotlib import plot_intermediate_values
            import matplotlib.pyplot as plt
            fig = plot_intermediate_values(self._study)
            if not legend:
                fig.get_legend().remove()
            plt.show()
    # ------------------------------------------------------------------------------------------

    def _set_optuna_verbosity(self, value):
        '''Set optimizaiton console output verbosity based on optuna.logging options'''
        value_to_optuna_verbosity_dict = {0 : 'CRITICAL',
                                          1 : 'CRITICAL',
                                          2 : 'ERROR',
                                          3 : 'WARNING',
                                          4 : 'INFO',
                                          5 : 'DEBUG'}
        optuna.logging.set_verbosity(getattr(optuna.logging, value_to_optuna_verbosity_dict[value]))
    # ------------------------------------------------------------------------------------------
        
    @timer
    def fit(self, X, y):
        '''
        Find optimized parameters for LightGBM model based on training data and 
        metric (defined by user at __init__).
        
        Considering the lower-better strategy:
            Regression optimization metric
                is selected based on eval_metric (passed by user at __init__), except for r2.
                If eval_metric == 'r2', then optimization metric is 'mean_squared_error'.
            Classification optimization metric
                Allways log_loss        
    
        LGB Classifier/Regressor is inferred based on metric (defined by user at __init__) 
        and target variable statistics.
        
        Initial LGBM parameters are inferred based on data stats & built in logic and are accesable
        by self._init_params
        
        Param_grid for hyperparameters search is inferred based on data stats & built in logic.
        
        Parameters
        ----------
        X : pd.DataFrame
            train features.
        y : pd.Series
            train labels.

        Returns
        -------
        None.

        '''
        validate_features_argument(X)
        validate_target_argument(y)

        optimization_metric_func = get_optimization_metric_func(self.metric)
        
        self.printer.print('Initiating LGBMTuner.fit', order=1)
        self.printer.print('Settings:', order=3)
        self.printer.print(f'Trying {self.trials} trials', order=4)
        self.printer.print(f'Evaluation metric: {self.metric} ', order=4)
        self.printer.print(f'Study direction: minimize {optimization_metric_func.__name__}', order=4)#{get_study_direction(metric)} {optimization_metric_func.__name__})
        print()
            
        self.target_classes = y.unique().tolist()
        self._get_target_minimum(y)
        self._init_params_on_input(len(X), y)
        self._set_optuna_verbosity(self.verbosity)
        
        sampler = optuna.samplers.TPESampler(seed=self.seed)
        study = optuna.create_study(pruner=optuna.pruners.MedianPruner(n_warmup_steps=5), 
                                    sampler=sampler, 
                                    direction = 'minimize')#get_study_direction(metric))

        optimization_function = partial(self._objective, X = X.values, y = y.values)

        study.optimize(optimization_function, n_trials = self.trials)


        # populate the learned params into the suggested params
        temp_params = self._populate_best_params_to_init_params(study.best_params)

        # tune n_estimators    
        iteration, best_score = self.optimize_n_estimators(X.values, y.values, temp_params)
        temp_params['n_estimators'] = iteration
        self._best_params = temp_params        
        if self.refit:
            self.fit_optimized(X.values, y.values)
            self._save_feature_importances(X.columns)
        self._study = study

        if self.visualization:
            self.plot_optimization_history()
            self.plot_param_importances()
            self.plot_intermediate_values()
            if self.refit:
                self.plot_importances()
        
        if self.best_params['objective'] == 'multiclass':
            n_rounds_eval_metric = 'multi_logloss'
        else:
            n_rounds_eval_metric = self.metric
        # clean up
        self.eval_results_callback = None
        # --------------------------------
        break_symbol = '|'
        print()
        self.printer.print(f"Optuna hyperparameters optimization finished", order=3)
        self.printer.print(f"Best trial number:{study.best_trial.number:>2}{break_symbol:>5}     {optimization_metric_func.__name__}:{study.best_trial.value:>29}", order=4, breakline='-')
        self.printer.print(f'n_estimators optimization finished', order=3)
        self.printer.print(f'best iteration:{iteration:>5}{break_symbol:>4}     {n_rounds_eval_metric}:{best_score:>29}', order=4, breakline='=')



import unittest
import sys
sys.path.append('../../')
import pandas as pd

from verstack.stratified_continuous_split import scsplit

class Test_scsplit(unittest.TestCase):
        
    def test_scsplit(self):
        df = pd.read_parquet('/Users/danil/Library/CloudStorage/OneDrive-Personal/Neuton/datasets/_traditional/boston (rmse,ID,medv)/boston_train.parquet')        
        train, test = scsplit(df, stratify=df['medv'], test_size = 0.5)        
        percent_diff_in_mean_of_column_used_for_stratification = (train['medv'].mean() - test['medv'].mean()) / train['medv'].mean()
        result = percent_diff_in_mean_of_column_used_for_stratification<0.05
        self.assertTrue(result)
        
if __name__ == '__main__':
    unittest.main()


import unittest
import sys
sys.path.append('../../')

from verstack.tools import Printer

class TestPrinter(unittest.TestCase):
        
    def test_Printer(self):

        result = None
        try:
            printer = Printer(verbose=True)
            printer.print('Message printed successfully', order=0)
        except:
            result='error'
        
        self.assertIsNone(result)
        
if __name__ == '__main__':
    unittest.main()


import pandas as pd
import copy
from sklearn.model_selection import KFold, StratifiedKFold
from verstack.stacking.optimise_params import optimise_params
from verstack.stacking.generate_default_layers import generate_default_layers
from verstack.stacking.args_validators import *
from verstack.tools import timer, Printer

'''
TODO: 
    1. Add scaling pipeline for linear models
'''

class Stacker:
    
    __version__ = '0.0.5'
    
    def __init__(self, 
                 objective, 
                 auto = False, 
                 num_auto_layers = 2, 
                 metafeats = True, 
                 epochs = 200, 
                 gridsearch_iterations = 10,
                 stacking_feats_depth = 1,
                 include_X = False,
                 verbose = True):
        '''
        Automatic stacking ensemble configuration, training, and features population to train/test sets.
        
        If stacking more than 1 layer, subsequent layers will can be trained either 
            on the predictions of the previous layer/layers including/not including 
                metafeats and/or original X features; based on the init settings.
        
        Additional statistical meta features can be created from the stacked predictions:
            - pairwise differences between the stacked predictions are created for all pairs (recursively)
            - mean and std for all the stacked features in a layer are created as two extra meta feats
            
        Includes manual and automatic stacking layers creation. 
        - Manual mode reuires user to define models instances, add them to Stacker layers via .add_layer() method.
            Layers must be added after Stacker initialisation separately.
            After creating stacking features using the fit_transform() method, user can add additional layers and 
                perform fit_transform()/transform() on the train/test sets (which have to include stacked features 
                from the previous iteration again to update the Stacker instance and add new features to the train/test sets.
        
        - Automatic mode constructs 2 layers:
            layer_1: 14 models
                - LGBM(max_depth = 12)
                - XGB(max_depth = 10, n_jobs = -1)
                - GradientBoosting(max_depth = 7)
                - kerasModel(num_layers = 3)
                - kerasModel(num_layers = 2)
                - kerasModel(num_layers = 1)
                - ExtraTree(max_depth = 12)
                - RandomForest(max_depth = 7)
                - Linear/LogisticRegression()
                - KNeighbors(n_neighbors=15)
                - KNeighbors(n_neighbors=10)
                - SVR(kernel = 'rbf')
                - DecisionTree(max_depth = 15)
                - DecisionTree(max_depth = 8)

            layer_2: two models
                - LGBM(max_depth = 3)
                - Ridge()
            
            Automatic mode will performs gridsearch of hyperparameters (others than the defined by default)
            on all the models in the default list.

                 num_auto_layers = 2, 
                 metafeats = True, 
                 epochs = 200, 
                 gridsearch_iterations = 10):
        
        Parameters
        ----------
        objective : str
            task type: 'regression', 'binary', 'multiclass'.
        auto : bool, optional
            automated construction of two layers with 14 and 2 models accordingly. The default is False.
        auto_num_layers : int, optional
            Number of layers to create in auto mode. Can except either 1 or 2. The default is 2.
        metafeats : bool, optional
            Flag to create statistical meta features for each layer. The default is True.
        epochs : int, optional
            Number of epochs for the 3 neural networks defined in the auto mode. The default is 200.
        gridsearch_iteration : int
            Number of gridsearch iterations for optimising hyperparameters of models in auto mode. The default is 10.
        stacking_feats_depth : int, optinal
            Defines the features used by subsequent layers to train the stacking models. 
            Can take values between 1 and 4 where:
                1 = use predictions from one previous layer
                2 = use predictions from one previous layer and meta features
                3 = use predictions from all previous layers
                4 = use predictions from all previous layers and meta features
                The default is 1.
        include_X : bool
            Flag to use original X features for subsequent layer training. The default is False.
        verbose : bool
            Flag to print stacking progress to the console. The default is True.

        Returns
        -------
        None.

        '''
        self.verbose = verbose
        self.printer = Printer(verbose=self.verbose)
        self.objective = objective
        self.auto = auto
        self.num_auto_layers = num_auto_layers
        self.metafeats = metafeats
        self.layers = {}
        self.trained_models = {}
        self._trained_models_list_buffer = None
        self._extra_layers_for_test_set_application = []
        self.epochs = epochs
        self.gridsearch_iterations = gridsearch_iterations
        self.stacking_feats_depth = stacking_feats_depth
        self.include_X = include_X
        self.stacked_features = {} #lists of stacked features by layers

        self._set_default_layers()

    # print init parameters when calling the class instance
    def __repr__(self):
        return f'Stacker(objective: {self.objective}\
            \n        auto: {self.auto}\
            \n        num_auto_layers: {self.num_auto_layers}\
            \n        metafeats: {self.metafeats}\
            \n        epochs : {self.epochs}\
            \n        gridsearch_iterations: {self.gridsearch_iterations}\
            \n        stacking_feats_depth: {self.stacking_feats_depth}\
            \n        include_X: {self.include_X}\
            \n        verbose : {self.verbose})'

    
    # VALIDATE INIT ARGUMENTS
    # -------------------------------------------------------------------------
    # objective
    @property
    def objective(self):
        return self._objective

    @objective.setter
    def objective(self, value):
        validate_objective(value)
        self._objective = value
    # -------------------------------------------------------------------------
    # auto
    @property
    def auto(self):
        return self._auto

    @auto.setter
    def auto(self, value):
        validate_bool_arg(value)
        self._auto = value
    # -------------------------------------------------------------------------
    # num_auto_layers
    @property
    def num_auto_layers(self):
        return self._num_auto_layers

    @num_auto_layers.setter
    def num_auto_layers(self, value):
        validate_num_auto_layers(value)
        self._num_auto_layers = value
    # -------------------------------------------------------------------------
    # metafeats
    @property
    def metafeats(self):
        return self._metafeats

    @metafeats.setter
    def metafeats(self, value):
        validate_bool_arg(value)
        self._metafeats = value
    # -------------------------------------------------------------------------
    # epochs
    @property
    def epochs(self):
        return self._epochs

    @epochs.setter
    def epochs(self, value):
        validate_epochs(value)
        self._epochs = value
    # -------------------------------------------------------------------------
    # gridsearch_iterations
    @property
    def gridsearch_iterations(self):
        return self._gridsearch_iterations

    @gridsearch_iterations.setter
    def gridsearch_iterations(self, value):
        validate_gridsearch_iterations(value)
        self._gridsearch_iterations = value
    # -------------------------------------------------------------------------
    # include_X
    @property
    def include_X(self):
        return self._include_X

    @include_X.setter
    def include_X(self, value):
        validate_bool_arg(value)
        self._include_X = value
    # -------------------------------------------------------------------------
    # verbose
    @property
    def verbose(self):
        return self._verbose

    @verbose.setter
    def verbose(self, value):
        validate_bool_arg(value)
        self._verbose = value
    # =========================================================================

    def _set_default_layers(self):
        '''Enable or not enable default layers based on self.auto init attribute'''
        if not self.auto:
            return
        else:
            layers = self._get_default_layers()
            for layer in layers:
                self.add_layer(layer)

    def _get_default_layers(self):
        '''Configure 1 or 2 default layers'''
        layers = []        
        layer_1, layer_2 = generate_default_layers(self.objective, self.epochs, self.verbose)
        layers.append(layer_1)
        if self.num_auto_layers == 2:
            layers.append(layer_2)
        return layers

    def _create_feat_name(self, layer):
        '''Assign feature name based on layer and consecutive number'''
        feat_count_in_layer = len(self.trained_models[layer].keys())        
        feat_name = '_'.join([layer, str(feat_count_in_layer)])
        return feat_name

    # -------------------------------------------------------------------------   
    def _predict_class_proba_or_value(self, model, X):
        '''Create single vector with predictions: 
            - classes for multiclass
            - positive class probabilities for binary
            - predicted numeric values for regression
            
        '''
        if self.objective == 'binary':
            try:
                pred = model.predict_proba(X)[:,1].flatten()
            except:
                pred = model.predict(X).flatten()
        else:
            pred = model.predict(X).flatten()
        return pred

    def _predict_by_model(self, model, X):
        '''Create predictions by passed model.'''
        pred = self._predict_class_proba_or_value(model, X)
        pred_series = pd.Series(pred)
        return pred_series
    # -------------------------------------------------------------------------            
    def _train_predict(self, model, X_train, y_train, X_test):
        '''Fit configured model, predict outputs by _predict_class_proba_or_value() method'''
        model.fit(X_train, y_train)
        pred = self._predict_class_proba_or_value(model, X_test)
        return pred, model

    def _configure_kfold_splitter(self):
        '''Initialize stratified or random kfold iterator based on objective'''
        if self.objective == 'regression':
            kfold = KFold(n_splits = 4, shuffle = True)
        else:
            kfold = StratifiedKFold(n_splits = 4, shuffle = True)
        return kfold

    def _train_predict_by_model(self, model, X, y):
        '''Train/predict by model on folds, create prediction series from each fold prediction
            Optimize model hyperparameters if self.auto
            Save list of trained model instances into self._trained_models_list_buffer
            for latter moving them into self.trained_models storage.
                
        '''
        # create placeholder for stacked feat
        pred_series = pd.Series(index = X.index)
        trained_models_list = []
        kfold = self._configure_kfold_splitter()

        if self.gridsearch_iterations > 0:
            model = optimise_params(model, X, y, self.objective, self.gridsearch_iterations, self.verbose)

        fold = 0
        for train_ix, test_ix in kfold.split(X,y):
            X_train = X.loc[train_ix, :]
            y_train = y.loc[train_ix]
            X_test = X.loc[test_ix, :]
            # create independent model instance for each fold
            fold_model = copy.deepcopy(model)
            pred, model = self._train_predict(fold_model, X_train, y_train, X_test)
            pred_series.loc[test_ix] = pred.flatten()
            trained_models_list.append(fold_model)            
            fold+=1
            if fold%2 == 0:
                self.printer.print(f'fold {fold} trained/predicted', 4)
        self._trained_models_list_buffer = trained_models_list
        return pred_series

    def _get_stack_feat(self, model, X, y = None):
        '''Apply stacking features creatin to either train or test set'''
        if isinstance(y, pd.Series):
            new_feat = self._train_predict_by_model(model, X, y)
        else:
            new_feat = self._predict_by_model(model, X)
        return new_feat

    # -------------------------------------------------------------------------
    def _create_layer_name(self):
        '''Create layer name from next consecutive number'''
        layer_keys = list(self.layers.keys())
        new_layer_ix = len(layer_keys) + 1
        new_layer_name = 'layer_' + str(new_layer_ix)
        return new_layer_name
    
    def _increment_layer(self, models_list):
        '''Increment layer with models to self.layers'''
        if not self.layers:
            self.layers['layer_1'] = models_list
        else:
            new_layer = self._create_layer_name()
            self.layers[new_layer] = models_list
            
    def add_layer(self, models_list):
        '''
        Add layer to Stacker, requires list of models instances as argument.

        Models list is added to Stacker.layers
        
        Parameters
        ----------
        models_list : list
            list of models instances each of which must contain methods:
                - fit()
                - predict()
                - predict_proba() for classification tasks

        Returns
        -------
        None.

        '''        
        validate_models_list(models_list)
        self._increment_layer(models_list)
    # -------------------------------------------------------------------------
    def _get_stacked_layer_feats(self, X, layer, metafeats):
        '''Get list of features names created within layer'''
        if metafeats:
            layer_feats = self.stacked_features[layer]
        else:
            layer_feats = list(self.trained_models[layer].keys())
        return layer_feats
            
    def _get_original_feats(self, X):
        '''Get names of original X feats (without any created stacking features)'''
        stacked_feats = list(self.stacked_features.values())
        if type(stacked_feats[0]) == list: # flatten nested list
            stacked_feats = sum(stacked_feats, [])
        original_feats = [col for col in X if col not in stacked_feats]
        return original_feats

    def _get_applicable_feats(self, X, layer):
        '''Get list of applicable features for trainin/prediction:
            - for layer_1  : all featues in X
            - for layer_2+ : preceeding layer outputs
                or preceeding layers outputs 
                    can include metafeats self.metafeats
                    can include original X features if self.include_X
            
        '''

        layer_number = int(layer.split('_')[1])
        
        if layer_number > 1:          
            # get last layer for creating applicable_feats based on a single preceeding layer
            preceeding_layers_list = list(self.stacked_features.keys())
            if layer in preceeding_layers_list:
                preceeding_layers_list = preceeding_layers_list[: preceeding_layers_list.index(layer)]
            last_layer = [l for l in preceeding_layers_list if l != layer][-1]
            
            generate_metafeats_flag = self.metafeats
            
            if self.stacking_feats_depth == 1:
                applicable_feats = self._get_stacked_layer_feats(X, last_layer, metafeats = generate_metafeats_flag)
            elif self.stacking_feats_depth == 2:
                applicable_feats = self._get_stacked_layer_feats(X, last_layer, metafeats = generate_metafeats_flag)
            elif self.stacking_feats_depth == 3:
                applicable_feats = []
                for l in preceeding_layers_list:
                    applicable_feats += self._get_stacked_layer_feats(X, l, metafeats = generate_metafeats_flag)
            else: # self.stacking_feats_depth == 4:
                applicable_feats = []
                for l in preceeding_layers_list:
                    applicable_feats += self._get_stacked_layer_feats(X, l, metafeats = generate_metafeats_flag)

            if self.include_X:
                # include original X features
                applicable_feats = applicable_feats + self._get_original_feats(X)
        else:
            applicable_feats = X.columns.tolist()
        return applicable_feats
    # -------------------------------------------------------------------------

    def _extract_most_common_multiclass_pred(self, preds_lists):
        '''Extract single (most) common prediction from list of lists with predictions from multiple models
            applicable for multiclass predictions
            
        '''
        def most_common_of_all_preds(lst):
            return max(set(lst), key=lst.count)
        
        # create list of tuples with n predictions from n models for each value
        preds_tuples = list(list(zip(*preds_lists)))
        result = [most_common_of_all_preds(pred) for pred in preds_tuples]
        return result                    

    def _average_predictions(self, preds_from_models, models_cnt):
        '''Create singe predictions vector from predictions from multiple models'''
        if self.objective == 'multiclass':
            preds_from_models = pd.Series(self._extract_most_common_multiclass_pred(preds_from_models))
        else:
            preds_from_models = preds_from_models/models_cnt
        return preds_from_models

    def _store_preds_together(self, preds_from_models, pred):
        '''Store predictions from multiple models into a single object.
            Append multiclass predictions to list of lists
            or sum binary probabilities/regression predictions in a series
        
        '''
        if self.objective == 'multiclass':
            preds_from_models.append(list(pred))                       
        else:                        
            if type(preds_from_models) != pd.Series:
                preds_from_models = pred
            else:
                preds_from_models += pred
        return preds_from_models
    # -------------------------------------------------------------------------

    def _create_new_feats_in_test(self, X, y, layer, applicable_feats):
        '''Header function to create stacking feats in train set by models in layer'''
        new_feats = []

        for feat_name in self.trained_models[layer].keys():                      
            models_list = self.trained_models[layer][feat_name]
            # create placeholder for predicting with all models for feat
            preds_from_models = []
            models_cnt = 0
            for model in models_list:
                pred = self._get_stack_feat(model, X[applicable_feats], y)
                preds_from_models = self._store_preds_together(preds_from_models, pred)
                models_cnt+=1
            preds_from_models = self._average_predictions(preds_from_models, models_cnt)
            preds_from_models.name = feat_name
            new_feats.append(preds_from_models)                
            self.printer.print(f'predicted with model {len(new_feats)}', 3)
        return new_feats
    
    def _create_new_feats_in_train(self, X, y, layer, applicable_feats):
        '''Header function to create stacking feats in test set by models in layer'''        
        new_feats = []
        for model in self.layers[layer]:
            feat_name = self._create_feat_name(layer)
            new_feat = self._get_stack_feat(model, X[applicable_feats], y)
            # append trained models from buffer to self.trained_models_list for layer/feature
            self.trained_models[layer][feat_name] = self._trained_models_list_buffer
            # clean up
            self._trained_models_list_buffer = None
            new_feats.append(pd.Series(new_feat, name = feat_name))
        return new_feats
    # -------------------------------------------------------------------------
        
    def _apply_single_layer(self, layer, X, y = None):
        '''Create stacked feats from all models in layer for train or test set
            train and test set is distinguished by y being None or not

        '''
        if y is None:
            self.printer.print(f'Predicting with {layer} models', 2)
        else:
            self.printer.print(f'Training/predicting with {layer} models', 2)
        cols_before_layer_stacking = X.columns.tolist()
        # create layer placeholder in self.trained_models list
        if layer not in self.trained_models.keys():
            self.trained_models[layer] = {}
        # get list of feature names to train/predict by models in layer
        applicable_feats = self._get_applicable_feats(X, layer)


        # create stacked feats in test set
        if y is None:
            new_feats = self._create_new_feats_in_test(X, y, layer, applicable_feats)
        # ---------------------------------------------------------------------
        # create stacked feats in train set
        else:
            new_feats = self._create_new_feats_in_train(X, y, layer, applicable_feats)
        for feat in new_feats:
            X = pd.concat([X, feat], axis = 1) 
        # ---------------------------------------------------------------------            
        # add metafeats to layer
        if self.metafeats:
            X = self._create_stacking_meta_features(X, layer)            
        cols_after_layer_stacking = [col for col in X if col not in cols_before_layer_stacking]
        if layer not in self.stacked_features.keys():
            self.stacked_features[layer] = cols_after_layer_stacking
        return X
    # -------------------------------------------------------------------------
    def _get_features_pairs_recursive(self, layer):
        '''Create list of all possible pairs combination in list'''
        feats = layer.copy()
        pairs = []
        for ix in range(len(feats)):
            feat = feats.pop(0)
            for feat2 in feats:
                pairs.append([feat, feat2])
        return pairs
    
    def _get_diff(self, feat1, feat2):
        '''Subtract pd.Series(feat2_ from pd.Series(feat1)'''
        return feat1 - feat2
    # -------------------------------------------------------------------------
    def _create_stacking_meta_features(self, X, layer):
        '''Generate statistics from stacked predictions; not applicable for multiclass'''

        if self.objective == 'multiclass':
            return X
        else:
            features_names_in_layer = list(self.trained_models[layer].keys())
            features_pairs = self._get_features_pairs_recursive(features_names_in_layer)
            metafeats_df = pd.DataFrame()        
            for pair in features_pairs:
                feat1 = pair[0]
                feat2 = pair[1]
                meta_feat_name = f'diff_{feat1}_{feat2}'
                diff = self._get_diff(X[feat1], X[feat2])            
                metafeats_df[meta_feat_name] = diff
            metafeats_df[f'{layer}_std'] = X[features_names_in_layer].std(axis = 1)
            metafeats_df[f'{layer}_mean'] = X[features_names_in_layer].mean(axis = 1)
            return pd.concat([X, metafeats_df], axis = 1)
        
    def _apply_all_layers(self, X, y = None):
        '''Wrapper for calling the self._apply_single_layer() method on all predefined layers'''
        for layer in list(self.layers.keys()):
            X = self._apply_single_layer(layer, X, y)
        return X

    def _apply_all_or_extra_layers_to_train(self, X, y = None):
        '''Perform initial stacking (on train set) on predefined layers or additional stacking with layers added 
            after calling fit_transform() method on initially predefined layers
            
        '''
        # search for layers added after fit_transform()
        layers_added_after_fit_transform = [x for x in self.layers.keys() if x not in self.trained_models.keys()]
        # append layers added after fit transform to buffer for test set transform()
        self._extra_layers_for_test_set_application += layers_added_after_fit_transform

        # apply extra layers on train set
        if layers_added_after_fit_transform:
            for layer in layers_added_after_fit_transform:
                X = self._apply_single_layer(layer, X, y)
        else:
            # if no extra layers apply all layers on train set
            X = self._apply_all_layers(X, y)      
        return X
    
    def _apply_all_or_extra_layers_to_test(self, X):
        '''Perform initial stacking (on train set) on predefined layers or additional stacking with layers added 
            after calling fit_transform()/transform() methods on initially predefined layers
        
        '''
        # iterate over self._extra_layers_for_test_set_application list
        if self._extra_layers_for_test_set_application:
            for ix in range(len(self._extra_layers_for_test_set_application)):
                layer = self._extra_layers_for_test_set_application.pop(0)
                X = self._apply_single_layer(layer, X)
        else:
            # if no extra layers apply all layers on test set
            X = self._apply_all_layers(X)          
        return X

    @timer
    def fit_transform(self, X, y):
        '''
        Train/predict/append to X stacking features from models defined in self.layers.

        Parameters
        ----------
        X : pd.DataFrame - train features
        y : pd.Series - train target.

        Returns
        -------
        X_with_stacked_feats : pd.DataFrame
            train featues with appended stacking features.

        '''
        X.reset_index(drop=True, inplace=True)
        self.printer.print('Initiating Stacker.fit_transform', order=1)
        validate_fit_transform_args(X, y)
        X_with_stacked_feats = X.reset_index(drop=True).copy()
        X_with_stacked_feats = self._apply_all_or_extra_layers_to_train(X_with_stacked_feats, y)
        return X_with_stacked_feats
    
    def transform(self, X):
        '''Create stacking features on the test set from models saved in self.trained_models
        Parameters
        ----------
        X : pd.DataFrame - test features

        Returns
        -------
        X_with_stacked_feats : pd.DataFrame
            test featues with appended stacking features.

        '''
        self.printer.print('Initiating Stacker.transform', order=1)
        validate_transform_args(X)
        X_with_stacked_feats = X.reset_index(drop=True).copy()
        X_with_stacked_feats = self._apply_all_or_extra_layers_to_test(X_with_stacked_feats)
        return X_with_stacked_feats


from setuptools import setup, find_packages
import os

# parse __version__ from version.py
exec(open('verstack/version.py').read())

# parse long_description from readme.md
with open("README.md", "r") as fh:
    long_description = fh.read()

# we conditionally add python-snappy based on the presence of an env var
dependencies = ['pandas', 'numpy']
rtd_build_env = os.environ.get('READTHEDOCS', False)
if not rtd_build_env:
    dependencies.append('xgboost')
    dependencies.append('scikit-learn==1.0.1')
    dependencies.append('lightgbm==3.3.0')
    dependencies.append('optuna==2.10.0')
    dependencies.append('plotly==5.3.1')
    dependencies.append('matplotlib')
    dependencies.append('python-dateutil==2.8.1')
    dependencies.append('holidays==0.11.3.1')
    dependencies.append('mlxtend')
    dependencies.append('tensorflow==2.7.0')
    dependencies.append('keras==2.7.0')
    dependencies.append('category_encoders==2.4.0')

setup(
  name = 'verstack',
  packages = find_packages(),
  version = __version__,
  license='MIT',
  description = "Machine learning tools to make a Data Scientist's work more efficient",
  long_description=long_description,
  long_description_content_type="text/markdown",
  author = 'Danil Zherebtsov',
  author_email = 'danil.com@me.com',
  url = 'https://github.com/DanilZherebtsov/verstack',
  download_url = 'https://github.com/DanilZherebtsov/verstack/archive/refs/tags/3.2.4.tar.gz',
  keywords = ['impute', 'missing', 'values', 'stratify', 'nan', 'continuous', 'multiprocessing', 'concurrent', 'timer'],
  install_requires=dependencies,
  classifiers=[
  'Development Status :: 4 - Beta',
  'Intended Audience :: Developers',
  'Topic :: Software Development :: Build Tools',
  'License :: OSI Approved :: MIT License',
  'Programming Language :: Python :: 3',
  'Programming Language :: Python :: 3.4',
  'Programming Language :: Python :: 3.5',
  'Programming Language :: Python :: 3.6',
  'Programming Language :: Python :: 3.7'
  ]
)



def timer(func):
    '''
    Decorator to print func execution time

    Parameters
    ----------
    func : function to decorate

    Returns
    -------
    wrapped func: function execution time result

    '''
    import time
    from functools import wraps
    @wraps(func)
    def wrapped(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        elapsed = round(end-start,5)
        if elapsed < 60:
            print(f"\nTime elapsed for {func.__name__} execution: {elapsed} seconds")
        elif 60 < elapsed < 3600:
            minutes = int(elapsed/60)
            seconds = round(elapsed%60,3)
            print(f"\nTime elapsed for {func.__name__} execution: {minutes} min {seconds} sec")
        else:
            hours = int(elapsed // 60 // 60)
            minutes = int(elapsed //60 % 60)
            seconds = int(elapsed % 60)
            print(f"\nTime elapsed for function {func.__name__} execution: {hours} hour(s) {minutes} min {seconds} sec")
        return result
    return wrapped

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr 27 17:51:24 2022

@author: danil
"""

class Printer:
    
    __version__ = '0.1.2'
    
    def __init__(self, verbose=True):
        self.verbose=verbose

    def print(self, 
              message=None, 
              order=1, 
              breakline=None, 
              force_print=False, 
              leading_blank_paragraph=False, 
              trailing_blank_paragraph=False):
        '''Output messages to the console based on seniority level (order).

        Logic:
            order=0 - program title print
            order=1 - major function title print
            order=2 - minor function title print
            order=3 - internal function first order results
            order=4 - internal function second order results
            order=5 - internal function third order results
            order='error' - error message print including traceback
        Parameters
        ----------
        message : str
            message to print
        order : int, optional
            order to tabulate the message print, can take values between 1 and 4. The default is 1.
        breakline : str, optional
            String symbol to print a breakline
        force_print : bool, optional
            If True will print message even if self.verbose == False. 
                Applicable for non-error important messages that need to be printed.

        Returns
        -------
        None.

        '''
        import traceback
        
        leading_blank = '\n' if leading_blank_paragraph else ''
        trailing_blank = '\n' if trailing_blank_paragraph else ''

        message_prefix = {
            1       :"\n * ",
            2       :"\n   - ",
            3       :"     . ",
            4       :"     .. ",
            5       :"     ... ",
            'error' : f"{traceback.format_exc()}\n! "
            }
        
        if order!='error' and not self.verbose and not force_print:
            return

        if not message:
            if breakline:
                print(f' {breakline*75}')
        else:
            if order == 0:
                print('\n')
                print('-'*75)
                print(f'{message}')
                print('-'*75)
            else:
                print(f'{leading_blank}{message_prefix[order]}{message}{trailing_blank}')
                if breakline: 
                    print(f' {breakline*75}')


from verstack.tools import Printer
from sklearn.model_selection import RandomizedSearchCV

def get_grid_search_instance(model, params, objective, gridsearch_iterations):
    '''Create RandomisedGridSearch instance based on model, params, objective, gridsearch_iterations'''
    if objective == 'regression':
        scoring = 'neg_mean_squared_error'
    else:
        scoring = 'neg_log_loss'
    rand_gridsearch = RandomizedSearchCV(model,
                                         param_distributions=params,
                                         scoring=scoring,
                                         cv=3,
                                         n_iter = gridsearch_iterations)
    return rand_gridsearch
# -----------------------------------------------------------------------------
def optimise_params(model, X, y, objective, gridsearch_iterations, verbose):
    '''
    Optimise hyperparameters of model if model is supported by the predefined models types:
        - lightgbm.sklearn.LGBMRegressor / lightgbm.sklearn.LGBMClassifier
        - xgboost.sklearn.XGBRegressor / xgboost.sklearn.XGBClassifier
        - sklearn.ensemble.GradientBoostingRegressor / sklearn.ensemble.GradientBoostingClassifier
        - sklearn.tree._classes.ExtraTreeRegressor / sklearn.tree._classes.ExtraTreeClassifier
        - sklearn.ensemble._forest.RandomForestRegressor / sklearn.ensemble._forest.RandomForestClassifier
        - sklearn.linear_model._logistic.LogisticRegression
        - sklearn.linear_model._ridge.Ridge
        - sklearn.neighbors._regression.KNeighborsRegressor / sklearn.neighbors._classification.KNeighborsClassifier
        - sklearn.svm._classes.SVR / sklearn.svm._classes.SVC
        - sklearn.tree._classes.DecisionTreeRegressor / sklearn.tree._classes.DecisionTreeClassifier

    If model not supported, it is returned not optimised.
    
    Parameters
    ----------
    model : model class instance
        model to optimise.
    X : pd.DataFrame
        train features.
    y : pd.Series
        train labels.
    objective : str
        regression/binary/multiclass.
    gridsearch_iterations : int
        number of gridsearch iterations.

    Returns
    -------
    model : TYPE
        DESCRIPTION.

    '''
    import sklearn
    import lightgbm
    import xgboost
    printer = Printer(verbose=verbose)
    printer.print(f'Optimising model hyperparameters', 3)
    try:
        # LGBM OPTIMIZATION
        if type(model) in [lightgbm.sklearn.LGBMRegressor, lightgbm.sklearn.LGBMClassifier]:
            params = {'colsample_bytree'  : [0.5, 0.7, 1],
                      'subsample'         : [0.5, 0.7, 1],
                      'num_leaves'        : [16, 32, 64, 128, 255],
                      'reg_alpha'         : [1e-8, 1e-4, 1, 10],
                      'reg_lambda'        : [1e-8, 1e-4, 1, 10]}
            rand_gridsearch = get_grid_search_instance(model, params, objective, gridsearch_iterations)
        # ---------------------------------------------------------------------
        # XGB OPTIMIZATION
        if type(model) in [xgboost.sklearn.XGBRegressor, xgboost.sklearn.XGBClassifier]:
            params = {'subsample'         : [0.5, 0.7, 1.0],
                      'colsample_bytree'  : [0.5, 0.7, 1.0],
                      'min_child_weight'  : [0.1, 1, 5, 10, 20],
                      'gamma'             : [0.5, 1, 1.5, 2, 5],
                      'eta'               : [0.005, 0.01, 0.1]}
            rand_gridsearch = get_grid_search_instance(model, params, objective, gridsearch_iterations)
        # ---------------------------------------------------------------------
        # GradientBoosting OPTIMIZATION
        if type(model) in [sklearn.ensemble.GradientBoostingRegressor, sklearn.ensemble.GradientBoostingClassifier]:
            params = {'subsample'         : [0.5, 0.7, 1.0],
                      'max_features'      : [0.5, 0.7, 1.0],
                      'min_samples_split' : [2, 5, 10, 20],
                      'min_samples_leaf'  : [1, 3, 5, 10],
                      'learning_rate'     : [0.005, 0.01, 0.1]}
            rand_gridsearch = get_grid_search_instance(model, params, objective, gridsearch_iterations)
        # ---------------------------------------------------------------------
        # ExtraTrees OPTIMIZATION
        if type(model) in [sklearn.tree._classes.ExtraTreeRegressor, sklearn.tree._classes.ExtraTreeClassifier]:
            params = {'max_features'      : [0.5, 0.7, 1.0],
                      'min_samples_split' : [2, 5, 10, 20],
                      'min_samples_leaf'  : [1, 3, 5, 10]}
            rand_gridsearch = get_grid_search_instance(model, params, objective, gridsearch_iterations)
        # ---------------------------------------------------------------------
        # RandomForest OPTIMIZATION
        if type(model) in [sklearn.ensemble._forest.RandomForestRegressor, sklearn.ensemble._forest.RandomForestClassifier]:
            params = {'max_samples'       : [0.5, 0.7, 1.0],
                      'min_samples_split' : [2, 5, 10, 20],
                      'min_samples_leaf'  : [1, 3, 5, 10]}
            rand_gridsearch = get_grid_search_instance(model, params, objective, gridsearch_iterations)
        # ---------------------------------------------------------------------
        # LogisticRegression OPTIMIZATION
        if type(model) == sklearn.linear_model._logistic.LogisticRegression:
            params = {'penalty'           : ['l1', 'l2'],
                      'C'                 : [0.5, 0.7, 1]}
            rand_gridsearch = get_grid_search_instance(model, params, objective, gridsearch_iterations)
        # ---------------------------------------------------------------------
        # Ridge OPTIMIZATION
        if type(model) == sklearn.linear_model._ridge.Ridge:
            params = {'alpha'             : [0.5, 0.7, 1]}
            rand_gridsearch = get_grid_search_instance(model, params, objective, gridsearch_iterations)
        # ---------------------------------------------------------------------
        # KNN OPTIMIZATION
        if type(model) in [sklearn.neighbors._regression.KNeighborsRegressor, sklearn.neighbors._classification.KNeighborsClassifier]:
            params = {'weights'           : ['uniform', 'distance'],
                      'leaf_size'         : [10, 15, 20, 30, 40, 50]}
            rand_gridsearch = get_grid_search_instance(model, params, objective, gridsearch_iterations)
        # ---------------------------------------------------------------------
        # SVM OPTIMIZATION
        if type(model) in [sklearn.svm._classes.SVR, sklearn.svm._classes.SVC]:
            params = {'C'                 : [0.5, 0.7, 1]}
            rand_gridsearch = get_grid_search_instance(model, params, objective, gridsearch_iterations)
        # ---------------------------------------------------------------------
        # DecisionTree OPTIMIZATION
        if type(model) in [sklearn.tree._classes.DecisionTreeRegressor, sklearn.tree._classes.DecisionTreeClassifier]:
            params = {'min_samples_split' : [2, 5, 10, 20],
                      'min_samples_leaf'  : [1, 3, 5, 10]}
            rand_gridsearch = get_grid_search_instance(model, params, objective, gridsearch_iterations)
        # ---------------------------------------------------------------------
        rand_gridsearch.fit(X, y)
        model.set_params(**rand_gridsearch.best_params_)

    except:
        printer.print(f'Model not in optimisation list {model}', order = 4)
    
    return model
# -----------------------------------------------------------------------------



import fileinput
import sys
from verstack.version import __version__ as current_version

files = ['verstack/version.py', 'README.md', 'docs/source/index.rst']
         
def update_version(file, new_version):
    for line in fileinput.input(file, inplace=1):
        if current_version in line:
            line = line.replace(current_version, new_version)
        sys.stdout.write(line)
    print(f'{new_version} is set in {file}')

if __name__ == '__main__':
    args = sys.argv[1:]
    for file in files:
        update_version(file, args[0])





import unittest
import sys
sys.path.append('../../')
import numpy as np

from verstack import ThreshTuner

class TestThreshTuner(unittest.TestCase):
        
    def test_ThreshTuner(self):
        np.random.seed(42)
        pred = np.random.uniform(0,1,500)
        true = np.random.randint(0,2,500)
        
        tuner = ThreshTuner(n_thresholds=10, min_threshold=0.4, max_threshold=0.7)
        tuner.fit(true, pred)
        print(tuner.min_threshold)
        result = tuner.best_predict_ratio()['threshold'].values[0] == 0.5333333333333333
        self.assertTrue(result)
        
if __name__ == '__main__':
    unittest.main()